# -*- coding: utf-8 -*-
"""
This module holds all static methods used in other modules. Static methods are general functions to perform
calculations. Most methods herein are not part of API of ospgrillage - these functions are only called and accessed in
ospgrillage for the purpose of its meshing and calculations. However, users may find some static methods to be useful
for their workflow.
"""

import logging
import numpy as np
from scipy.spatial import distance
from scipy.optimize import fsolve, root
import warnings

logger = logging.getLogger(__name__)

__all__ = [
    "diff",
    "find_circle",
    "find_dict_key",
    "find_min_x_dist",
    "get_distance",
    "get_line_func",
    "get_slope",
    "get_y_intcp",
    "intersection",
    "inv_line_func",
    "is_between",
    "line_func",
    "arc_func",
    "onSegment",
    "orientation",
    "check_intersect",
    "check_dict_same_keys",
    "check_point_in_grid",
    "check_points_direction",
    "calculate_area_given_vertices",
    "create_arc_points",
    "find_plane_centroid",
    "get_patch_centroid",
    "rotate_point_about_point",
    "select_segment_function",
    "solve_zeta_eta",
    "sort_list_into_four_groups",
    "sort_vertices",
    "x_intcp_two_lines",
    "line",
]


def diff(li1, li2):
    """
    Static method to determine the difference between two list. used in OspGrillage.set_member() function to check member
    assignment
    """
    return list(list(set(li1) - set(li2)) + list(set(li2) - set(li1)))


def find_circle(x1, y1, x2, y2, x3, y3):
    """
    Find the center and radius of a circle passing through three points.

    Parameters
    ----------
    x1, y1 : float
        Coordinates of the first point.
    x2, y2 : float
        Coordinates of the second point.
    x3, y3 : float
        Coordinates of the third point.

    Returns
    -------
    list
        A list containing [center, radius] where center is [h, k] coordinates
        of the circle center and radius is the circle radius.
    """
    x12 = x1 - x2
    x13 = x1 - x3

    y12 = y1 - y2
    y13 = y1 - y3

    y31 = y3 - y1
    y21 = y2 - y1

    x31 = x3 - x1
    x21 = x2 - x1

    # x1^2 - x3^2
    sx13 = pow(x1, 2) - pow(x3, 2)

    # y1^2 - y3^2
    sy13 = pow(y1, 2) - pow(y3, 2)

    sx21 = pow(x2, 2) - pow(x1, 2)
    sy21 = pow(y2, 2) - pow(y1, 2)

    f = ((sx13) * (x12) + (sy13) * (x12) + (sx21) * (x13) + (sy21) * (x13)) // (
        2 * ((y31) * (x12) - (y21) * (x13))
    )

    g = ((sx13) * (y12) + (sy13) * (y12) + (sx21) * (y13) + (sy21) * (y13)) // (
        2 * ((x31) * (y12) - (x21) * (y13))
    )

    c = -pow(x1, 2) - pow(y1, 2) - 2 * g * x1 - 2 * f * y1

    # eqn of circle be x^2 + y^2 + 2*g*x + 2*f*y + c = 0
    # where centre is (h = -g, k = -f) and
    # radius r as r^2 = h^2 + k^2 - c
    h = -g
    k = -f
    sqr_of_r = h * h + k * k - c

    # r is the radius
    r = round(np.sqrt(sqr_of_r), 5)

    logger.debug("Centre = (%s, %s)", h, k)
    logger.debug("Radius = %s", r)

    return [[h, k], r]


# -----------------------------------------------------------------------------------------------------------------------
# Sweep path equations - Add future functions here
# which the given three points lie


def line_func(m=None, c=None, x=None, h=None, v=None, R=None):
    """
    Compute y coordinate for a line or circular arc.

    For a straight line: y = m*x + c. For a circular arc (upper half):
    y = sqrt(R^2 - (x - h)^2) + v.

    Parameters
    ----------
    m : float, optional
        Slope of the line (used for straight line only).
    c : float, optional
        Y-intercept of the line (used for straight line only).
    x : float or list, optional
        X coordinate(s) at which to evaluate the function.
    h : float, optional
        X coordinate of arc center (used for arc only).
    v : float, optional
        Y coordinate of arc center (used for arc only).
    R : float, optional
        Radius of the arc (used for arc only).

    Returns
    -------
    float or ndarray
        Y coordinate(s) corresponding to the input x value(s).

    Raises
    ------
    ValueError
        If arguments are insufficient to determine line or arc function.
    """
    curve = False
    if not all([h is None, v is None, R is None]):
        curve = True

    if not curve:
        # straight line
        if isinstance(x, list):
            y = m * x[0] + c
        else:
            y = m * x + c
    elif curve:
        y = np.sqrt((R) ** 2 - (x - h) ** 2) + v
    else:
        raise ValueError(
            "line function missing arguments for valid function selection: check arguments"
        )

    return y


def inv_line_func(m, c, y):
    """
    Compute x coordinate on a line given y, slope, and y-intercept.

    Solves x = (y - c) / m for a line y = m*x + c.

    Parameters
    ----------
    m : float or None
        Slope of the line.
    c : float or None
        Y-intercept of the line.
    y : float
        Y coordinate.

    Returns
    -------
    float
        X coordinate. Returns 0 if slope is zero, None, or intercept is None.
    """
    x = (y - c) / m if all([m != 0, m is not None, c is not None]) else 0
    return x


def arc_func(h, v, R, x, r=0):
    """
    Compute y coordinate on a circular arc (upper half).

    Parameters
    ----------
    h : float
        X coordinate of the arc center.
    v : float
        Y coordinate of the arc center.
    R : float
        Radius of the arc.
    x : float or ndarray
        X coordinate(s) at which to evaluate the arc.
    r : float, optional
        Additional radius offset added to R (default 0).

    Returns
    -------
    float or ndarray
        Y coordinate(s) on the circular arc.
    """
    y = np.sqrt((R + r) ** 2 - (x - h) ** 2) + v
    return y


def create_arc_points(point1, radius, length, num_inc):
    """
    Create points along a circular arc.

    Generates evenly spaced points along an arc of given arc length and radius.
    The arc is centered at [point1.x, -radius] with center point in the xz-plane
    (y = 0).

    Parameters
    ----------
    point1 : object
        Object with .x attribute specifying the x coordinate of the arc start.
    radius : float
        Radius of the circular arc.
    length : float
        Arc length along the radius.
    num_inc : int
        Number of points to generate along the arc.

    Returns
    -------
    tuple
        Two lists: (x_curve, z_curve) containing the x and z coordinates of
        points along the arc.
    """
    # function to create points along arc `length` of a sector with `angle`
    start_angle = np.pi / 2  # 90 degrees
    angle = length / radius  # calculate angle of sector
    end_angle = (
        start_angle - angle
    )  # difference is the end point's angle (about center_point)
    center_point = [point1.x, -radius]  # x z , model plane = 0 default
    # find point2
    angle_inc = np.linspace(start_angle, end_angle, num_inc)
    # point2 = [center_point[0] + radius*np.cos(end_angle), center_point[1] + radius * np.sin(end_angle)]

    x_curve = [center_point[0] + radius * np.cos(ang) for ang in angle_inc]
    z_curve = [center_point[1] + radius * np.sin(ang) for ang in angle_inc]

    return x_curve, z_curve


# -----------------------------------------------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------------------------------------------
def select_segment_function(curve_flag, d, x, r=0, m=0, c=0):
    """
    Evaluate a line or arc function based on a flag.

    Parameters
    ----------
    curve_flag : bool
        If True, evaluate as circular arc. If False, evaluate as straight line.
    d : list or tuple
        For curve_flag=True: [center, radius] where center is [h, k].
        Unused for curve_flag=False.
    x : float or ndarray
        X coordinate(s) at which to evaluate the function.
    r : float, optional
        Additional radius offset for arc (default 0).
    m : float, optional
        Slope for line (default 0).
    c : float, optional
        Y-intercept for line (default 0).

    Returns
    -------
    float or ndarray
        Y coordinate(s) evaluated at x.
    """
    if curve_flag:
        y = arc_func(h=d[0][0], v=d[0][1], R=d[1], x=x, r=r)
    else:
        y = line_func(m, c, x)
    return y


def find_dict_key(my_dict, key):
    """
    Find and parse the dictionary key corresponding to a given value.

    Parameters
    ----------
    my_dict : dict
        Dictionary to search.
    key : object
        Value to search for in the dictionary.

    Returns
    -------
    object
        The parsed key corresponding to the given value, evaluated from its
        string representation.

    Raises
    ------
    ValueError
        If the key value is not found in the dictionary.
    """
    import ast
    import numpy as _np

    raw = list(my_dict.keys())[list(my_dict.values()).index(key)]
    try:
        return ast.literal_eval(raw)
    except ValueError:
        # Keys may contain numpy repr like np.float64(...) in newer numpy versions
        return eval(raw, {"np": _np})


def x_intcp_two_lines(m1, m2, c1, c2):
    """
    Find the x coordinate of intersection of two lines.

    Computes the x coordinate where two lines y = m1*x + c1 and y = m2*x + c2
    intersect.

    Parameters
    ----------
    m1, m2 : float
        Slopes of the first and second lines.
    c1, c2 : float
        Y-intercepts of the first and second lines.

    Returns
    -------
    float
        X coordinate of intersection.
    """
    x = (c2 - c1) / (m1 - m2)
    return x


def line(p1, p2):
    """
    Convert two points to line equation coefficients.

    Computes coefficients A, B, C of the line equation Ax + By + C = 0 passing
    through two points.

    Parameters
    ----------
    p1, p2 : tuple or list
        Points with coordinates (x, y).

    Returns
    -------
    tuple
        Coefficients (A, B, C) of the line equation Ax + By + C = 0.
    """
    A = p1[1] - p2[1]
    B = p2[0] - p1[0]
    C = p1[0] * p2[1] - p2[0] * p1[1]
    return A, B, -C


def intersection(L1, L2):
    """
    Find the intersection point of two lines.

    Parameters
    ----------
    L1, L2 : tuple
        Line coefficients (A, B, C) for equations Ax + By + C = 0.

    Returns
    -------
    tuple or False
        (x, y) coordinates of intersection if lines are not parallel, else False.
    """
    D = L1[0] * L2[1] - L1[1] * L2[0]
    Dx = L1[2] * L2[1] - L1[1] * L2[2]
    Dy = L1[0] * L2[2] - L1[2] * L2[0]
    if D != 0:
        x = Dx / D
        y = Dy / D
        return x, y
    else:
        return False


def get_y_intcp(m, x, y):
    """
    Calculate y-intercept of a line given slope and a point.

    Parameters
    ----------
    m : float or None
        Slope of the line.
    x, y : float
        Coordinates of a point on the line.

    Returns
    -------
    float or None
        Y-intercept c (from y = m*x + c), or None if slope is None.
    """
    c = y - x * m if m is not None else None
    return c


def get_line_func(skew_angle, node_point):
    """
    Compute line function (slope and intercept) from skew angle and a point.

    Parameters
    ----------
    skew_angle : float
        Angle in degrees.
    node_point : list or tuple
        Coordinates of a point on the line. If length < 3, uses indices [0, 1];
        otherwise uses [0, 2].

    Returns
    -------
    tuple
        (m, c) slope and y-intercept of the line y = m*x + c.
    """
    m = 1 / np.tan(-skew_angle / 180 * np.pi)
    if len(node_point) < 3:
        c = get_y_intcp(m=m, x=node_point[0], y=node_point[1])
    else:
        c = get_y_intcp(m=m, x=node_point[0], y=node_point[2])
    return m, c


def find_min_x_dist(const_point, ref_point):
    """
    Compute pairwise distances between two sets of points.

    Parameters
    ----------
    const_point : ndarray
        Points on a construction line.
    ref_point : ndarray
        Points on an arbitrary line.

    Returns
    -------
    ndarray
        Pairwise distance matrix between const_point and ref_point.
    """
    # constant point is node point on construction line
    # ref_point is roving point on the arbitrary line
    d = distance.cdist(const_point, ref_point)
    return d


def get_slope(pt1, pt2):
    """
    Calculate slope and angle of line between two points.

    Computes slope in the xz-plane (y = 0) and the angle (phi) from horizontal.

    Parameters
    ----------
    pt1, pt2 : array-like
        Points in 3D with coordinates [x, y, z]. Only x and z are used.

    Returns
    -------
    tuple
        (m, phi) where m is slope (dz/dx) or None if vertical line, and phi is
        the angle in radians from horizontal.
    """
    # pt must be [x y z]
    # function claculates the slope for two points inthe 2-D plane, y= 0
    if (pt1[0] - pt2[0]) == 0:
        m = None
    else:
        m = (pt1[2] - pt2[2]) / (pt1[0] - pt2[0])

    # find phi angle
    if m is None:
        phi = np.pi / 2
    else:
        phi = np.arctan(m)
    return m, phi


def solve_zeta_eta(xp, zp, x1, z1, x2, z2, x3, z3, x4, z4):
    """
    Solve for natural coordinates (eta, zeta) from global coordinates.

    Maps a global point (xp, zp) to isoparametric coordinates (eta, zeta) on a
    quadrilateral element with corners at (x1, z1), (x2, z2), (x3, z3), (x4, z4).
    Uses bilinear mapping and numerical root finding.

    Parameters
    ----------
    xp, zp : float
        Global coordinates of the point to map.
    x1, z1, x2, z2, x3, z3, x4, z4 : float
        Global coordinates of element corners.

    Returns
    -------
    tuple
        (eta, zeta) natural coordinates in range [-1, 1] for bilinear elements.
    """
    # create function to solve for eta and zeta - dynamically for varying parameters x1-x4, z1-z4, zp,xp

    # mapping of natural coordinate eta{-1:1}, zeta{-1:1} to global coordinate (x,z)
    def obj_func(x, xp, zp, x1, x2, x3, x4, z1, z2, z3, z4):
        eta = 4 * zp - (
            (1 - x[0]) * (1 - x[1]) * z1
            + (1 + x[0]) * (1 - x[1]) * z2
            + (1 + x[0]) * (1 + x[1]) * z3
            + (1 - x[0]) * (1 + x[1]) * z4
        )
        zeta = 4 * xp - (
            (1 - x[0]) * (1 - x[1]) * x1
            + (1 + x[0]) * (1 - x[1]) * x2
            + (1 + x[0]) * (1 + x[1]) * x3
            + (1 - x[0]) * (1 + x[1]) * x4
        )
        return eta, zeta

    # if initial xp zp is very close (order 1e-8) to the centre (0,0,0) of natural coordinate,
    # avoid fsolve RunTimeWarning by
    if all([np.isclose(xp, (x1 + x2) / 2), np.isclose(zp, (z1 + z3) / 2)]):
        root_func = [0, 0]

    else:
        root_func = fsolve(
            obj_func,
            np.array([-0.4444444444444, 0.99999999]),
            args=(xp, zp, x1, x2, x3, x4, z1, z2, z3, z4),
        )  # here the initial values are chosen to avoid RunTimeWarning where initial guess is very close
        # to the final solution of eta and zeta

    eta = root_func[0]
    zeta = root_func[1]
    return eta, zeta


def get_distance(a, b):
    """
    Calculate Euclidean distance between two points in the xz-plane.

    Parameters
    ----------
    a, b : object
        Points with .x and .z attributes.

    Returns
    -------
    float
        Euclidean distance between the two points.
    """
    return np.sqrt((a.x - b.x) ** 2 + (a.z - b.z) ** 2)


def is_between(a, c, b):
    """
    Check if point c lies on the line segment between points a and b.

    Parameters
    ----------
    a, b, c : object
        Points with .x and .z attributes.

    Returns
    -------
    bool
        True if c is on the line segment from a to b, False otherwise.
    """
    return get_distance(a, c) + get_distance(c, b) == get_distance(a, b)


def calculate_area_given_vertices(p_list):
    """
    Calculate area of a polygon given sorted vertices.

    Uses the Shoelace formula to compute the area. Vertices must be sorted
    counter-clockwise (typically via sort_vertices function).

    Parameters
    ----------
    p_list : list
        List of point namedtuples (LoadVertex or Point) with .x, .z attributes,
        sorted counter-clockwise.

    Returns
    -------
    float
        Area of the polygon in the xz-plane.
    """
    # input list of namedtuple LoadVertex or Point
    # note: p_list must have been sorted in counter clockwise via sort_vertices function
    # (this is called prior to this function)
    x_list = []
    y_list = []
    z_list = []
    A_mag = 0
    for point in p_list:
        x_list.append(point.x)
        y_list.append(point.y)
        z_list.append(point.z)
    # add first point to list
    x_list.append(p_list[0].x)
    y_list.append(p_list[0].y)
    z_list.append(p_list[0].z)
    for count in range(len(x_list) - 1):
        A_mag += x_list[count] * z_list[count + 1] - x_list[count + 1] * z_list[count]
    A = np.abs(A_mag / 2)
    return A


def check_point_in_grid(inside_point, point_list):
    """
    Check if a point is inside a polygon.

    Uses the sign of cross products to determine if a point is inside a polygon
    defined by the given vertices.

    Parameters
    ----------
    inside_point : object
        Point with .x, .z attributes to test.
    point_list : list
        List of point namedtuples defining polygon vertices.

    Returns
    -------
    bool
        True if inside_point is inside the polygon, False otherwise.
    """
    # ref: solution 3 https://www.eecs.umich.edu/courses/eecs380/HANDOUTS/PROJ2/InsidePoly.html
    # check counter clockwise and to the left (greater than 0)
    # put points in list
    pt0 = point_list  # list of point tuples
    # rotate point
    pt1 = pt0[1:] + [pt0[0]]
    inside = True
    if inside_point.z is None:
        inside = False
        return inside
    # check if point is clockwise via sign of signed_area
    signed_area = check_points_direction(
        point_list
    )  # sign area < 0 means points are clockwise, and vice versa
    for count, point0 in enumerate(pt0):
        side = (inside_point.z - point0.z) * (pt1[count].x - point0.x) - (
            inside_point.x - point0.x
        ) * (pt1[count].z - point0.z)
        # check if point is outside
        if any(
            [
                side
                < 0
                <= signed_area,  # side > 0 means left, side < 0 means right, side = 0 means on the line path
                side > 0 > signed_area,
            ]
        ):  # nodes are counterclockwise and point to right (outside)
            inside = False  # nodes are clockwise and point to the left (outside) these condition
    return inside  # return line = outside


def check_points_direction(point_list):
    """
    Calculate signed area to determine polygon winding order.

    Computes the signed area of a polygon. Positive indicates counter-clockwise
    ordering, negative indicates clockwise.

    Parameters
    ----------
    point_list : list
        List of point namedtuples with .x and .z attributes.

    Returns
    -------
    float
        Signed area of the polygon. Positive for counter-clockwise, negative
        for clockwise ordering.
    """
    # list of point tuples
    # ref http://mathworld.wolfram.com/PolygonArea.html
    signed_area = 0
    last_point = None
    first_point = None
    x2 = None
    z2 = None
    for counter, point in enumerate(point_list):
        x1 = point.x
        z1 = point.z
        if first_point is None:
            # save point as first point
            first_point = point

        if counter == len(point_list) - 1:
            x2 = first_point.x
            z2 = first_point.z
        else:
            x2 = point_list[counter + 1].x
            z2 = point_list[counter + 1].z
        signed_area += x1 * z2 - x2 * z1

        last_point = point  # set current point as last point
    return signed_area


# ----------------------------------------------------------------------------------------------------------
# function to check intersection of line segments
def onSegment(p, q, r):
    """
    Check if point q lies on segment pr (given collinear points).

    Parameters
    ----------
    p, q, r : object
        Points with .x and .z attributes (note: .y is mistakenly used for .z
        in this implementation).

    Returns
    -------
    bool
        True if q is within the bounding box of segment pr, False otherwise.
    """
    # point nameTuple p, q and r
    if (
        (q.x <= max(p.x, r.x))
        and (q.x >= min(p.x, r.x))
        and (q.y <= max(p.z, r.z))
        and (q.z >= min(p.z, r.z))
    ):
        return True
    return False


def orientation(p, q, r):
    """
    Find the orientation of an ordered triplet of points.

    Parameters
    ----------
    p, q, r : object
        Points with .x and .z attributes.

    Returns
    -------
    int
        0 if collinear, 1 if clockwise, 2 if counter-clockwise.
    """
    # to find the orientation of an ordered triplet (p,q,r)
    # function returns the following values:
    # 0 : Colinear points
    # 1 : Clockwise points
    # 2 : Counterclockwise

    # See https://www.geeksforgeeks.org/orientation-3-ordered-points/amp/
    # for details of below formula.
    # modified - p, q, r must be point namedtuple
    val = (float(q.z - p.z) * (r.x - q.x)) - (float(q.x - p.x) * (r.z - q.z))
    if val > 0:
        # Clockwise orientation
        return 1
    elif val < 0:
        # Counterclockwise orientation
        return 2
    else:
        # Colinear orientation
        return 0


def check_intersect(p1, q1, p2, q2):
    """
    Check if two line segments intersect.

    Determines if line segments p1-q1 and p2-q2 intersect, including special
    cases where segments are collinear.

    Parameters
    ----------
    p1, q1, p2, q2 : object
        Points defining two line segments, with .x and .z attributes.

    Returns
    -------
    tuple
        (general_intersect, colinear) where general_intersect is True if
        segments intersect, and colinear is True if intersection is collinear.
    """
    # ref https://www.geeksforgeeks.org/check-if-two-given-line-segments-intersect/
    general_intersect = False
    colinear = False
    # Find the 4 orientations required for
    # the general and special cases
    o1 = orientation(p1, q1, p2)
    o2 = orientation(p1, q1, q2)
    o3 = orientation(p2, q2, p1)
    o4 = orientation(p2, q2, q1)

    # General case
    if (o1 != o2) and (o3 != o4):
        general_intersect = True
        colinear = False
        return general_intersect, colinear

    # Special Cases
    # p1 , q1 and p2 are colinear and p2 lies on segment p1q1
    if (o1 == 0) and onSegment(p1, p2, q1):
        general_intersect = True
        colinear = True
        return general_intersect, colinear

    # p1 , q1 and q2 are colinear and q2 lies on segment p1q1
    if (o2 == 0) and onSegment(p1, q2, q1):
        general_intersect = True
        colinear = True
        return general_intersect, colinear

    # p2 , q2 and p1 are colinear and p1 lies on segment p2q2
    if (o3 == 0) and onSegment(p2, p1, q2):
        general_intersect = True
        colinear = True
        return general_intersect, colinear

    # p2 , q2 and q1 are colinear and q1 lies on segment p2q2
    if (o4 == 0) and onSegment(p2, q1, q2):
        general_intersect = True
        colinear = True
        return general_intersect, colinear

    # If none of the cases
    return general_intersect, colinear


# --------------------------------------------------------------------------------------------
def find_plane_centroid(point_list):
    """
    Function to find centroid given a set of vertices (for patch load). function calculates 2D plane x z - y plane is
    the model plane.
    :param point_list: list of Point() namedtuples
    :type point_list: list
    """
    x = 0
    z = 0
    for point in point_list:
        x += point.x
        z += point.z
    x_c = x / len(point_list)
    z_c = z / len(point_list)
    return [x_c, z_c]


def sort_vertices(point_list, node_tag_list=None):
    """
    Sort polygon vertices counter-clockwise by angle from centroid.

    Arranges points in counter-clockwise order suitable for isoparametric
    element mapping and area calculations.

    Parameters
    ----------
    point_list : list
        List of point namedtuples with .x and .z attributes.
    node_tag_list : list, optional
        List of node tags corresponding to points (default None).

    Returns
    -------
    tuple
        (sorted_points, sorted_node_tag) where sorted_points are the vertices
        in counter-clockwise order and sorted_node_tag are the corresponding tags.
    """
    # note the node_tag_list must correspond to that of point_list, this is ensure in the higher level
    # functions which calls sort_vertices
    if node_tag_list is None:
        node_tag_list = []
    center_x_z = find_plane_centroid(point_list)
    angle_list = []
    for point in point_list:
        # calculate angle
        x = np.array(point.x - center_x_z[0])
        z = np.array(point.z - center_x_z[1])
        angle_list.append(np.arctan2(z, x))
        # angle_list.append(np.arctan((point.z - center_x_z[1]) / (point.x - center_x_z[0])))
    # Angles are intentionally left in the [-π, π] range from arctan2.
    #
    # For typical rectangular grids, a bottom-left vertex sits in the 3rd
    # quadrant relative to the centroid, giving the most-negative angle and
    # therefore sorting first.  solve_zeta_eta() (and hermite_shape_function_2d)
    # expect node 1 to be at the bottom-left corner of the element, so this
    # implicit convention must be preserved.
    #
    # Normalising to [0, 2π] would instead start at the top-right vertex
    # (smallest positive angle), which rotates the isoparametric mapping and
    # corrupts the load distribution.  Any future normalisation refactor must
    # also make solve_zeta_eta rotation-invariant.
    # sort for counter clockwise
    sorted_points = [x for _, x in sorted(zip(angle_list, point_list))]

    sorted_node_tag = [x for _, x in sorted(zip(angle_list, node_tag_list))]
    return sorted_points, sorted_node_tag


def get_patch_centroid(point_list):
    """
    Calculate weighted centroid of points with per-node load values.

    Finds the centroid of a polygon accounting for distributed area loads at
    each vertex (weighted centroid).

    Parameters
    ----------
    point_list : list
        List of point namedtuples with .x, .y, .z, and .p (load) attributes.

    Returns
    -------
    tuple
        (xc, yc, zc) coordinates of the weighted centroid.
    """
    # function to find centroid on a plane (x-z) accounting for separate value of area load on each node point
    m_total = []
    mx = 0
    my = 0
    mz = 0
    for point in point_list:
        m_total.append(point.p)
        mx += point.x * point.p
        mz += point.z * point.p
        my += point.y * point.p
    sum_m = sum(m_total) if sum(m_total) > 0 else 1
    xc = mx / sum_m
    zc = mz / sum_m
    yc = my / sum_m
    return xc, yc, zc


def check_dict_same_keys(d_1, d_2):
    """
    Merge two nested dictionaries, combining values for common keys.

    For grids present in both dictionaries, combines the value lists while
    removing duplicates.

    Parameters
    ----------
    d_1, d_2 : dict
        Dictionaries with grid numbers as keys and nested dicts as values.

    Returns
    -------
    dict
        Merged dictionary with combined values for common keys.
    """
    merged = {**d_1, **d_2}
    # function to check if two dicts have same key
    same_key = [k in d_2 for k in d_1.keys()]
    if any(same_key):
        common_key = [k for (k, v) in zip(d_1, same_key) if v]
        # get val from d_1 and d_2
        for grid in common_key:  # key is grid number
            first_list = d_1[grid]
            second_list = d_2[grid]
            updated_list = dict()
            for key, val_1 in first_list.items():
                # if key == 'ends':
                #     val_end = val_1 if val_1 else second_list.get(key)
                #     updated_list.setdefault(key, val_end)
                # else:
                combine_val = []
                val_2 = second_list.get(key)
                for val in val_1 + val_2:
                    if val not in combine_val:
                        combine_val.append(val)

                updated_list.setdefault(key, combine_val)
            merged.update({grid: updated_list})

    return merged


def sort_list_into_four_groups(group_list: list, option: str = None):
    """
    Function to sort a list of group number into four groups , returns a dict with keys [0,1,2,3] placing the ordered
    group_list into each key. Purpose is to distinguish main grillage elements based on the group positioning
    e.g. if group_list = [0,1,2,3,4,5,6], returns {0:[0,6], 1:[1],2:[2,3,4],3:[5]}
    :param option: type of group sorting - default None for beam, "shell" for shell model
    :type option: str
    :param group_list:
    :returns: dict with keys [0,1,2,3] and values be the sorted group list as shown in description
    """
    # sort asccending
    output_dict = dict()
    if option == "shell":
        # add two proxy element at start and end of group list
        group_list = [group_list[0] - 1] + group_list + [group_list[-1] + 1]

    # edges
    output_dict[0] = [group_list[0], group_list[-1]]
    # instantiate
    exterior_beam_group_1 = [group_list[1]]
    exterior_beam_group_2 = [group_list[-2]]
    interior_beam_group = group_list[2 : len(group_list) - 2]

    # check for special cases
    if len(group_list) > 2:
        if len(group_list) == 3:  # 2 edge and 1 interior
            interior_beam_group = [group_list[1]]
            exterior_beam_group_1 = []
            exterior_beam_group_2 = []
        elif len(group_list) == 4:  # 2 edge, ext_1 and ext_2
            interior_beam_group = []

    output_dict[1] = exterior_beam_group_1
    output_dict[2] = interior_beam_group
    output_dict[3] = exterior_beam_group_2
    return output_dict


def rotate_point_about_point(center_x, center_y, angle, point: list):
    """
    Rotates point about a second point at [center_x, center_y] by angle in the coordinate system

    param point: list containing the x and y coordinate
    type point: list
    param angle: Angle in radians
    type angle: int or float
    returns: list of coordinate for the rotated point
    """

    s = np.sin(angle)  # rad
    c = np.cos(angle)  # rad
    if isinstance(point, list):
        x_or = point[0] - center_x
        y_or = point[1] - center_y

        x_rot = x_or * c - y_or * s
        y_rot = x_or * s + y_or * c

        # translate back to origin and store to var
        rotated_point = [x_rot + center_x, y_rot + center_y]
    else:
        raise ValueError("point= requires a valid list or tuple")

    return rotated_point



"""

"""

from fixtures import *

sys.path.insert(0, os.path.abspath("../"))

# Adopted units: N and m
kilo = 1e3
milli = 1e-3
N = 1
m = 1
mm = milli * m
m2 = m**2
m3 = m**3
m4 = m**4
kN = kilo * N
MPa = N / ((mm) ** 2)
GPa = kilo * MPa


# =====================================================================================================================
# Tests
# =====================================================================================================================
def test_inputs():
    with pytest.raises(ValueError) as e_info:
        vertex_errors = og.create_load_vertex(x=3, y=0, z=3)
        vertex_errors = og.create_load_vertex()


# test to check compound load position relative to global are correct
def test_compound_load_positions():
    #
    location = og.create_load_vertex(x=5, z=-2, p=20)  # create load point
    Single = og.create_load(loadtype="point", name="single point", point1=location)
    # front_wheel = PointLoad(name="front wheel", localpoint1=LoadPoint(2, 0, 2, 50))
    # Line load
    barrierpoint_1 = og.create_load_vertex(x=-1, z=0, p=2)
    barrierpoint_2 = og.create_load_vertex(x=11, z=0, p=2)
    Barrier = og.LineLoading(point1=barrierpoint_1, point2=barrierpoint_2)

    M1600 = og.create_compound_load(name="Lane and Barrier")
    M1600.add_load(Single)
    M1600.add_load(load=Barrier)  # this overwrites the current global pos of line load
    # the expected midpoint (reference point initial is 6,0,0) is now at 9,0,5 (6+3, 0+0, 5+0)
    # when setting the global coordinate, the global coordinate is added with respect to ref point (9,0,5)
    # therefore (3+4, 0+0, 3+5) = (13,0,8)
    # M1600.set_global_coord(og.Point(4, 0, 3))
    a = 2
    # check if point Single is same as point Single's load vertex
    assert M1600.compound_load_obj_list[0].load_point_1 == og.LoadPoint(
        x=5, y=0, z=-2, p=20
    )
    assert M1600.compound_load_obj_list[1].load_point_1 == og.LoadPoint(
        x=-1, y=0, z=0, p=2
    )
    assert M1600.compound_load_obj_list[1].load_point_2 == og.LoadPoint(
        x=11.0, y=0, z=0, p=2
    )

    # now we set global and see if correct
    M1600.set_global_coord(og.Point(4, 0, 3))
    assert M1600.compound_load_obj_list[0].load_point_1 == og.LoadPoint(
        x=9, y=0, z=1, p=20
    )
    assert M1600.compound_load_obj_list[1].load_point_1 == og.LoadPoint(
        x=3, y=0, z=3, p=2
    )
    assert M1600.compound_load_obj_list[1].load_point_2 == og.LoadPoint(
        x=15.0, y=0, z=3, p=2
    )


def test_point_load_getter(
    bridge_model_42_negative,
):  # test get_point_load_nodes() function
    # test if setter and getter is correct              # and assign_point_to_node() function

    example_bridge = bridge_model_42_negative
    # create reference point load
    location = og.create_load_vertex(x=5, y=0, z=2, p=20)
    Single = og.create_load(
        loadtype="point", name="single point", point1=location, shape_function="hermite"
    )
    ULS_DL = og.create_load_case(name="Point")
    ULS_DL.add_load(Single)  # ch
    example_bridge.add_load_case(ULS_DL)
    og.ops.wipe()

    #
    # assert example_bridge.load_case_list[0]["load_command"] == [
    #     "ops.load(12, *[0, 0.6075807082987842, 0, 0.37389582049155967, 0, 0.34166943132701877])\n",
    #     "ops.load(17, *[0, 1.4724192917012129, 0, 0.9061041795084391, 0, -0.613915767971676])\n",
    #     "ops.load(18, *[0, 12.685458513118162, 0, -3.6244167180337583, 0, -5.289120462525217])\n",
    #     "ops.load(13, *[0, 5.234541486881841, 0, -1.4955832819662394, 0, 2.943613562202012])\n",
    # ]
    # reference_command = example_bridge.load_case_list[0]["load_command"]
    #
    # assert reference_command
    def parse_load_call(call):
        # load_command entries are now (func_name, args, kwargs) tuples.
        # args = (node_tag, Fx, Fy, Fz, Mx, My, Mz)
        _name, args, _kwargs = call
        assert _name == "load", f"Unexpected func name: {_name}"
        node = args[0]
        # non-zero components are Fy, Mx, Mz (indices 2, 4, 6 in args)
        values = [float(args[i]) for i in (2, 4, 6) if args[i] != 0]
        return node, values

    expected_cmds = [
        (12, [0.6075807082987873, 0.3738958204915613, 0.34166943132702027]),
        (17, [1.4724192917012136, 0.9061041795084389, -0.6139157679716769]),
        (18, [12.685458513118142, -3.624416718033756, -5.289120462525214]),
        (13, [5.234541486881858, -1.4955832819662453, 2.94361356220202]),
    ]

    for call, (exp_node, exp_vals) in zip(
        example_bridge.load_case_list[0]["load_command"], expected_cmds
    ):
        node, values = parse_load_call(call)
        assert node == exp_node
        for a, b in zip(values, exp_vals):
            assert abs(a - b) < 1e-9  # tolerance for float comparison

    # assert example_bridge.load_case_list[0]["load_command"] == [
    #     "ops.load(12, *[0, np.float64(0.6075807082987873), 0, np.float64(0.3738958204915613), 0, np.float64(0.34166943132702027)])\n",
    #     "ops.load(17, *[0, np.float64(1.4724192917012136), 0, np.float64(0.9061041795084389), 0, np.float64(-0.6139157679716769)])\n",
    #     "ops.load(18, *[0, np.float64(12.685458513118142), 0, np.float64(-3.624416718033756), 0, np.float64(-5.289120462525214)])\n",
    #     "ops.load(13, *[0, np.float64(5.234541486881858), 0, np.float64(-1.4955832819662453), 0, np.float64(2.94361356220202)])\n",
    # ]


# test point load returning None when point (loadpoint) is outside of mesh
def test_point_load_outside_straight_mesh(bridge_model_42_negative):
    example_bridge = bridge_model_42_negative
    location = og.create_load_vertex(x=5, y=0, z=-2, p=20)
    Single = og.create_load(loadtype="point", name="single point", point1=location)
    ULS_DL = og.create_load_case(name="Point")
    ULS_DL.add_load(Single)  # ch
    example_bridge.add_load_case(ULS_DL)
    grid_nodes, _ = example_bridge._get_point_load_nodes(point=location)
    assert grid_nodes is None


# test general line with line load in the bounds of the mesh
def test_line_load(bridge_model_42_negative):
    example_bridge = bridge_model_42_negative
    # create reference line load
    barrierpoint_1 = og.create_load_vertex(x=3, y=0, z=3, p=2)
    barrierpoint_2 = og.create_load_vertex(x=10, y=0, z=3, p=2)
    Barrier = og.create_load(
        name="Barrier curb load",
        point1=barrierpoint_1,
        point2=barrierpoint_2,
    )
    ULS_DL = og.create_load_case(name="Barrier")
    ULS_DL.add_load(Barrier)  # ch
    example_bridge.add_load_case(ULS_DL)
    example_bridge.analyze()
    example_bridge.get_results()
    ref_answer = [
        {
            7: {
                "long_intersect": [],
                "trans_intersect": [[3.1514141550424406, 0, 3.0]],
                "edge_intersect": [],
                "ends": [[3, 0, 3]],
            },
            8: {
                "long_intersect": [],
                "trans_intersect": [
                    [3.1514141550424406, 0, 3.0],
                    [4.276919210414739, 0, 3.0],
                ],
                "edge_intersect": [],
                "ends": [],
            },
            11: {
                "long_intersect": [],
                "trans_intersect": [
                    [4.276919210414739, 0, 3.0],
                    [5.4024242657870385, 0, 3.0],
                ],
                "edge_intersect": [],
                "ends": [],
            },
            16: {
                "long_intersect": [],
                "trans_intersect": [
                    [5.4024242657870385, 0, 3.0],
                    [6.302828310084881, 0, 3.0],
                ],
                "edge_intersect": [],
                "ends": [],
            },
            22: {
                "long_intersect": [],
                "trans_intersect": [
                    [6.302828310084881, 0, 3.0],
                    [7.227121232563658, 0, 3.0],
                ],
                "edge_intersect": [],
                "ends": [],
            },
            31: {
                "long_intersect": [],
                "trans_intersect": [[10.0, 0, 3.0]],
                "edge_intersect": [],
                "ends": [[10, 0, 3]],
            },
            32: {
                "long_intersect": [],
                "trans_intersect": [[10.0, 0, 3.0], [9.075707077521221, 0, 3.0]],
                "edge_intersect": [],
                "ends": [[10, 0, 3]],
            },
            56: {
                "long_intersect": [],
                "trans_intersect": [
                    [7.227121232563658, 0, 3.0],
                    [8.15141415504244, 0, 3.0],
                ],
                "edge_intersect": [],
                "ends": [],
            },
            62: {
                "long_intersect": [],
                "trans_intersect": [
                    [8.15141415504244, 0, 3.0],
                    [9.075707077521221, 0, 3.0],
                ],
                "edge_intersect": [],
                "ends": [],
            },
        }
    ]
    for grid_key, intersect_dict in example_bridge.global_line_int_dict[0].items():
        assert grid_key in ref_answer[0].keys()  # check keys is correct

        for intersect_type, grid_val in intersect_dict.items():
            # check values
            if grid_val:
                assert grid_val[0] == pytest.approx(
                    ref_answer[0][grid_key][intersect_type][0]
                )


# test line load function with line load is vertical (slope = infinite) and start end points
def test_line_load_vertical_and_cross_outside_mesh(bridge_model_42_negative):
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative
    # create reference line load
    barrierpoint_1 = og.create_load_vertex(x=2, y=0, z=-3, p=2)
    barrierpoint_2 = og.create_load_vertex(x=2, y=0, z=8, p=2)
    Barrier = og.create_load(
        loadtype="line",
        name="Barrier curb load",
        point1=barrierpoint_1,
        point2=barrierpoint_2,
        shape_function="hermite",
    )
    ULS_DL = og.create_load_case(name="Barrier")
    ULS_DL.add_load(Barrier)  # ch
    example_bridge.add_load_case(ULS_DL)
    ref_ans = {
        1: {
            "long_intersect": [[2, 0, 0.0], [2, 0, 1.0]],
            "trans_intersect": [],
            "edge_intersect": [],
            "ends": [],
        },
        3: {
            "long_intersect": [[2, 0, 1.0]],
            "trans_intersect": [],
            "edge_intersect": [[2.0, 0, 2.2212250296583864]],
            "ends": [],
        },
    }
    for grid_key, intersect_dict in example_bridge.global_line_int_dict[0].items():
        assert grid_key in ref_ans.keys()  # check keys is correct

        for intersect_type, grid_val in intersect_dict.items():
            # check values
            if grid_val:
                assert grid_val[0] == pytest.approx(
                    ref_ans[grid_key][intersect_type][0]
                )


# test a line load which coincide with edge z = 0 or z = 7
def test_line_load_coincide_long_edge(bridge_model_42_negative):
    # when set line load z coordinate to z = 0 , test if line returns correct coincide node lines
    example_bridge = bridge_model_42_negative
    # create reference line load
    barrierpoint_1 = og.create_load_vertex(x=4, y=0, z=1, p=2)
    barrierpoint_2 = og.create_load_vertex(x=10, y=0, z=1, p=2)
    Barrier = og.create_load(
        loadtype="line",
        name="Barrier curb load",
        point1=barrierpoint_1,
        point2=barrierpoint_2,
    )
    ULS_DL = og.create_load_case(name="Barrier")
    ULS_DL.add_load(Barrier)  # ch
    example_bridge.add_load_case(ULS_DL)

    ref_answer = [
        {
            5: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[4, 0, 1]],
            },
            6: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[4, 0, 1]],
            },
            27: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[10, 0, 1]],
            },
            28: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[10, 0, 1]],
            },
            29: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[10, 0, 1]],
            },
            30: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[10, 0, 1]],
            },
        }
    ]
    for grid_key, intersect_dict in example_bridge.global_line_int_dict[0].items():
        assert grid_key in ref_answer[0].keys()  # check keys is correct

        for intersect_type, grid_val in intersect_dict.items():
            # check values
            if grid_val:
                assert grid_val[0] == pytest.approx(
                    ref_answer[0][grid_key][intersect_type][0]
                )


def test_line_load_coincide_transverse_member(bridge_42_0_angle_mesh):
    example_bridge = bridge_42_0_angle_mesh
    # og.opsplt.plot_model("nodes")

    # create reference line load

    barrierpoint_1 = og.create_load_vertex(x=7.5, y=0, z=1, p=2)
    # barrierpoint_1 = og.create_load_vertices(x=7.5, y=0, z=1, p=2)
    barrierpoint_2 = og.create_load_vertex(x=7.5, y=0, z=6, p=2)
    Barrier = og.create_load(
        loadtype="line",
        name="Barrier curb load",
        point1=barrierpoint_1,
        point2=barrierpoint_2,
        shape_function="hermite",
    )
    ULS_DL = og.create_load_case(name="Barrier")
    ULS_DL.add_load(Barrier)  # ch
    example_bridge.add_load_case(ULS_DL)
    ref_answer = [
        "ops.load(51, *[0, 0.0, 0, 0, 0, 0])\n",
        "ops.load(30, *[0, 0.0, 0, 0, 0, 0])\n",
        "ops.load(31, *[0, 0.0, 0, 0, 0, 0])\n",
        "ops.load(52, *[0, 0.0, 0, 0, 0, 0])\n",
        "ops.load(51, *[0, 1.2499999999999993, 0, 0, 0, 0])\n",
        "ops.load(30, *[0, 1.3877787807814452e-16, 0, 0, 0, 0])\n",
        "ops.load(31, *[0, 1.3877787807814462e-16, 0, 0, 0, 0])\n",
        "ops.load(52, *[0, 1.2500000000000002, 0, 0, 0, 0])\n",
        "ops.load(52, *[0, 1.2500000000000009, 0, 0, 0, 0])\n",
        "ops.load(31, *[0, -4.163336342344338e-16, 0, 0, 0, 0])\n",
        "ops.load(32, *[0, -4.1633363423443365e-16, 0, 0, 0, 0])\n",
        "ops.load(53, *[0, 1.2500000000000002, 0, 0, 0, 0])\n",
        "ops.load(53, *[0, 1.2500000000000009, 0, 0, 0, 0])\n",
        "ops.load(32, *[0, -4.163336342344338e-16, 0, 0, 0, 0])\n",
        "ops.load(33, *[0, -4.1633363423443365e-16, 0, 0, 0, 0])\n",
        "ops.load(54, *[0, 1.2500000000000002, 0, 0, 0, 0])\n",
        "ops.load(54, *[0, 1.2500000000000024, 0, 0, 0, 0])\n",
        "ops.load(33, *[0, -4.163336342344343e-16, 0, 0, 0, 0])\n",
        "ops.load(34, *[0, -4.163336342344331e-16, 0, 0, 0, 0])\n",
        "ops.load(55, *[0, 1.2499999999999987, 0, 0, 0, 0])\n",
        "ops.load(55, *[0, 0.0, 0, 0, 0, 0])\n",
        "ops.load(34, *[0, -0.0, 0, 0, 0, 0])\n",
        "ops.load(35, *[0, -0.0, 0, 0, 0, 0])\n",
        "ops.load(56, *[0, 0.0, 0, 0, 0, 0])\n",
    ]

    def _ref_vals(s):
        start = s.find("[")
        return eval(s[start : s.find("]") + 1])

    for i, load_command in enumerate(example_bridge.load_case_list[0]["load_command"]):
        _name, args, _kw = load_command
        pos = list(args[1:])  # drop node_tag; rest are the 6 load components
        pos_ref = _ref_vals(ref_answer[i])
        assert pos == pytest.approx(pos_ref)


def test_line_load_coincide_edge_beam(bridge_model_42_negative):
    # when set line load z coordinate to z = 0 , test if line returns correct coincide node lines

    example_bridge = bridge_model_42_negative
    # create reference line load
    barrierpoint_1 = og.create_load_vertex(x=5, y=0, z=1, p=2)
    barrierpoint_2 = og.create_load_vertex(x=10, y=0, z=1, p=2)
    Barrier = og.create_load(
        loadtype="line",
        name="Barrier curb load",
        point1=barrierpoint_1,
        point2=barrierpoint_2,
        shape_function="hermite",
    )
    ULS_DL = og.create_load_case(name="Barrier")
    ULS_DL.add_load(Barrier)  # ch
    example_bridge.add_load_case(ULS_DL)

    ref_ans = [
        {
            9: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[5, 0, 1]],
            },
            10: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[5, 0, 1]],
            },
            27: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[10, 0, 1]],
            },
            28: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[10, 0, 1]],
            },
            29: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[10, 0, 1]],
            },
            30: {
                "long_intersect": [],
                "trans_intersect": [],
                "edge_intersect": [],
                "ends": [[10, 0, 1]],
            },
        }
    ]

    for grid_key, intersect_dict in example_bridge.global_line_int_dict[0].items():
        assert grid_key in ref_ans[0].keys()  # check keys is correct

        for intersect_type, grid_val in intersect_dict.items():
            # check values
            if grid_val:
                assert grid_val[0] == pytest.approx(
                    ref_ans[0][grid_key][intersect_type][0]
                )


def test_line_load_outside_of_mesh(bridge_model_42_negative):
    # when set line load z coordinate to z = 0 , test if line returns correct coincide node lines
    example_bridge = bridge_model_42_negative
    # create reference line load
    barrierpoint_1 = og.create_load_vertex(x=3, y=0, z=-1, p=2)
    barrierpoint_2 = og.create_load_vertex(x=10, y=0, z=-1, p=2)
    Barrier = og.create_load(
        loadtype="line",
        point1=barrierpoint_1,
        point2=barrierpoint_2,
        shape_function="hermite",
    )
    ULS_DL = og.create_load_case(name="Barrier")
    ULS_DL.add_load(Barrier)  # ch
    example_bridge.add_load_case(ULS_DL)
    assert example_bridge.global_line_int_dict == [{}]


# test a default patch load - patch is within the mesh and sufficiently larger than a single grid
def test_patch_load(bridge_model_42_negative):
    example_bridge = bridge_model_42_negative

    lane_point_1 = og.create_load_vertex(x=5, z=3, p=5)
    lane_point_2 = og.create_load_vertex(x=8, z=3, p=5)
    lane_point_3 = og.create_load_vertex(x=8, z=5, p=5)
    lane_point_4 = og.create_load_vertex(x=5, z=5, p=5)
    Lane = og.create_load(
        loadtype="patch",
        point1=lane_point_1,
        point2=lane_point_2,
        point3=lane_point_3,
        point4=lane_point_4,
        shape_function="hermite",
    )
    ULS_DL = og.create_load_case(name="Lane")
    ULS_DL.add_load(Lane)  # ch
    example_bridge.add_load_case(ULS_DL)
    example_bridge.analyze()

    ref_answer = [
        "ops.load(19, *[0, 1.4068813192153762, 0, 0.7034406596076881, 0, 0.7034406596076881])\n",
        "ops.load(25, *[0, 1.4068813192153762, 0, 0.7034406596076881, 0, -0.7034406596076881])\n",
        "ops.load(26, *[0, 1.4068813192153762, 0, -0.7034406596076881, 0, -0.7034406596076881])\n",
        "ops.load(20, *[0, 1.4068813192153762, 0, -0.7034406596076881, 0, 0.7034406596076881])\n",
        "ops.load(25, *[0, 1.4442076913730961, 0, 0.7221038456865481, 0, 0.7221038456865481])\n",
        "ops.load(60, *[0, 1.4442076913730961, 0, 0.7221038456865481, 0, -0.7221038456865481])\n",
        "ops.load(61, *[0, 1.4442076913730961, 0, -0.7221038456865481, 0, -0.7221038456865481])\n",
        "ops.load(26, *[0, 1.4442076913730961, 0, -0.7221038456865481, 0, 0.7221038456865481])\n",
        "ops.load(13, *[0, 0.00883644799442324, 0, 0.0054378141504143, 0, 0.005492412044427657])\n",
        "ops.load(18, *[0, 0.09579386111020773, 0, 0.05895006837551243, 0, -0.025230076869609933])\n",
        "ops.load(19, *[0, 0.82530095725717, 0, -0.235800273502049, 0, -0.21736681610740755])\n",
        "ops.load(14, *[0, 0.07612939810579983, 0, -0.02175125660165713, 0, 0.04731924222891496])\n",
        "ops.load(18, *[0, 0.11705252575871955, 0, 0.0720323235438274, 0, 0.05852626287935976])\n",
        "ops.load(24, *[0, 0.11705252575871945, 0, 0.07203232354382733, 0, -0.05852626287935973])\n",
        "ops.load(25, *[0, 1.0084525296135802, 0, -0.28812929417530886, 0, -0.5042262648067901])\n",
        "ops.load(19, *[0, 1.0084525296135811, 0, -0.28812929417530914, 0, 0.5042262648067904])\n",
        "ops.load(24, *[0, 0.12015807992224165, 0, 0.07394343379830252, 0, 0.06007903996112076])\n",
        "ops.load(59, *[0, 0.12015807992224087, 0, 0.07394343379830204, 0, -0.06007903996112049])\n",
        "ops.load(60, *[0, 1.035208073176227, 0, -0.2957737351932079, 0, -0.517604036588114])\n",
        "ops.load(25, *[0, 1.0352080731762336, 0, -0.2957737351932098, 0, 0.5176040365881163])\n",
        "ops.load(59, *[0, 0.12494235320970869, 0, 0.07688760197520536, 0, 0.056897774410875845])\n",
        "ops.load(66, *[0, 0.07600612632374003, 0, 0.04677300081460926, 0, -0.040880164010709555])\n",
        "ops.load(67, *[0, 0.6548220114045306, 0, -0.1870920032584372, 0, -0.35219833609226747])\n",
        "ops.load(60, *[0, 1.0764264276528765, 0, -0.3075504079008217, 0, 0.490196210309085])\n",
        "ops.load(60, *[0, 1.501710976078226, 0, 0.750855488039113, 0, 0.6838674808999481])\n",
        "ops.load(67, *[0, 0.9135351721603328, 0, 0.4567675860801664, 0, -0.4913481251287184])\n",
        "ops.load(68, *[0, 0.9135351721603329, 0, -0.45676758608016643, 0, -0.49134812512871845])\n",
        "ops.load(61, *[0, 1.5017109760782261, 0, -0.7508554880391131, 0, 0.6838674808999482])\n",
        "ops.load(61, *[0, 0.5838652274992154, 0, 0.09731087124986965, 0, 0.2658876765739])\n",
        "ops.load(68, *[0, 0.35518247493593585, 0, 0.059197079155989554, 0, -0.1910361510500451])\n",
        "ops.load(69, *[0, 0.01023159392819577, 0, -0.006577453239554427, 0, -0.005503099001441673])\n",
        "ops.load(62, *[0, 0.016819162932076304, 0, -0.010812319027763346, 0, 0.007659315786079486])\n",
        "ops.load(15, *[0, 0.08992923272602986, 0, 0.0, 0, 0.0])\n",
        "ops.load(20, *[0, 0.36279806628438993, 0, 0.0, 0, 0.0])\n",
        "ops.load(21, *[0, 0.05030303322337963, 0, 0.0, 0, 0.0])\n",
        "ops.load(20, *[0, 0.5469954569109332, 0, 0.09116590948515592, 0, 0.2734977284554662])\n",
        "ops.load(26, *[0, 0.5469954569109291, 0, 0.09116590948515522, 0, -0.2734977284554649])\n",
        "ops.load(27, *[0, 0.01575707077521208, 0, -0.010129545498350628, 0, -0.007878535387606049])\n",
        "ops.load(21, *[0, 0.015757070775212195, 0, -0.010129545498350706, 0, 0.007878535387606089])\n",
        "ops.load(26, *[0, 0.5615079504058634, 0, 0.09358465840097817, 0, 0.2807539752029313])\n",
        "ops.load(61, *[0, 0.5615079504058589, 0, 0.09358465840097742, 0, -0.28075397520292983])\n",
        "ops.load(62, *[0, 0.01617512614337903, 0, -0.010398295377886516, 0, -0.008087563071689525])\n",
        "ops.load(27, *[0, 0.016175126143379157, 0, -0.0103982953778866, 0, 0.008087563071689568])\n",
        "ops.load(14, *[0, 0.10620730762527882, 0, 0.05310365381263941, 0, 0.06601456784167825])\n",
        "ops.load(19, *[0, 1.1513685229592219, 0, 0.575684261479611, 0, -0.3032461162212718])\n",
        "ops.load(20, *[0, 1.1513685229592223, 0, -0.5756842614796112, 0, -0.3032461162212719])\n",
        "ops.load(15, *[0, 0.10620730762527884, 0, -0.05310365381263942, 0, 0.06601456784167829])\n",
    ]

    def _ref_vals(s):
        start = s.find("[")
        return eval(s[start : s.find("]") + 1])

    for i, load_command in enumerate(example_bridge.load_case_list[0]["load_command"]):
        _name, args, _kw = load_command
        pos = list(args[1:])  # drop node_tag; rest are the 6 load components
        pos_ref = _ref_vals(ref_answer[i])
        assert pos == pytest.approx(pos_ref)  # check each pos


# test for patch load with linear shape function for load distribution
def test_patch_load_using_linear_shape_function(bridge_model_42_negative):
    example_bridge = bridge_model_42_negative

    lane_point_1 = og.create_load_vertex(x=5, y=0, z=3, p=5)
    lane_point_2 = og.create_load_vertex(x=8, y=0, z=3, p=5)
    lane_point_3 = og.create_load_vertex(x=8, y=0, z=5, p=5)
    lane_point_4 = og.create_load_vertex(x=5, y=0, z=5, p=5)
    Lane = og.create_load(
        loadtype="patch",
        point1=lane_point_1,
        point2=lane_point_2,
        point3=lane_point_3,
        point4=lane_point_4,
        shape_function="linear",
    )
    ULS_DL = og.LoadCase(name="Lane")
    ULS_DL.add_load(Lane)  # ch
    example_bridge.add_load_case(ULS_DL)
    example_bridge.analyze()
    ref_answer = [
        "ops.load(19, *[0, 1.4068813192153762, 0, 0, 0, 0])\n",
        "ops.load(25, *[0, 1.4068813192153762, 0, 0, 0, 0])\n",
        "ops.load(26, *[0, 1.4068813192153762, 0, 0, 0, 0])\n",
        "ops.load(20, *[0, 1.4068813192153762, 0, 0, 0, 0])\n",
        "ops.load(25, *[0, 1.4442076913730961, 0, 0, 0, 0])\n",
        "ops.load(60, *[0, 1.4442076913730961, 0, 0, 0, 0])\n",
        "ops.load(61, *[0, 1.4442076913730961, 0, 0, 0, 0])\n",
        "ops.load(26, *[0, 1.4442076913730961, 0, 0, 0, 0])\n",
        "ops.load(13, *[0, 0.03597169309041219, 0, 0, 0, 0])\n",
        "ops.load(18, *[0, 0.16524043980310846, 0, 0, 0, 0])\n",
        "ops.load(19, *[0, 0.6609617592124317, 0, 0, 0, 0])\n",
        "ops.load(14, *[0, 0.14388677236164832, 0, 0, 0, 0])\n",
        "ops.load(18, *[0, 0.22510101107446034, 0, 0, 0, 0])\n",
        "ops.load(24, *[0, 0.2251010110744602, 0, 0, 0, 0])\n",
        "ops.load(25, *[0, 0.9004040442978396, 0, 0, 0, 0])\n",
        "ops.load(19, *[0, 0.90040404429784, 0, 0, 0, 0])\n",
        "ops.load(24, *[0, 0.23107323061969504, 0, 0, 0, 0])\n",
        "ops.load(59, *[0, 0.231073230619694, 0, 0, 0, 0])\n",
        "ops.load(60, *[0, 0.924292922478775, 0, 0, 0, 0])\n",
        "ops.load(25, *[0, 0.9242929224787791, 0, 0, 0, 0])\n",
        "ops.load(59, *[0, 0.22487220771081853, 0, 0, 0, 0])\n",
        "ops.load(66, *[0, 0.16156717600735238, 0, 0, 0, 0])\n",
        "ops.load(67, *[0, 0.6462687040294102, 0, 0, 0, 0])\n",
        "ops.load(60, *[0, 0.899488830843275, 0, 0, 0, 0])\n",
        "ops.load(60, *[0, 1.4054512981926113, 0, 0, 0, 0])\n",
        "ops.load(67, *[0, 1.0097948500459475, 0, 0, 0, 0])\n",
        "ops.load(68, *[0, 1.0097948500459475, 0, 0, 0, 0])\n",
        "ops.load(61, *[0, 1.4054512981926115, 0, 0, 0, 0])\n",
        "ops.load(61, *[0, 0.5059624673493407, 0, 0, 0, 0])\n",
        "ops.load(68, *[0, 0.36352614601654004, 0, 0, 0, 0])\n",
        "ops.load(69, *[0, 0.040391794001838004, 0, 0, 0, 0])\n",
        "ops.load(62, *[0, 0.05621805192770482, 0, 0, 0, 0])\n",
        "ops.load(15, *[0, 0.08992923272602986, 0, 0, 0, 0])\n",
        "ops.load(20, *[0, 0.36279806628438993, 0, 0, 0, 0])\n",
        "ops.load(21, *[0, 0.05030303322337963, 0, 0, 0, 0])\n",
        "ops.load(20, *[0, 0.50647727491753, 0, 0, 0, 0])\n",
        "ops.load(26, *[0, 0.5064772749175274, 0, 0, 0, 0])\n",
        "ops.load(27, *[0, 0.05627525276861446, 0, 0, 0, 0])\n",
        "ops.load(21, *[0, 0.056275252768614746, 0, 0, 0, 0])\n",
        "ops.load(26, *[0, 0.5199147688943169, 0, 0, 0, 0])\n",
        "ops.load(61, *[0, 0.5199147688943141, 0, 0, 0, 0])\n",
        "ops.load(62, *[0, 0.05776830765492456, 0, 0, 0, 0])\n",
        "ops.load(27, *[0, 0.05776830765492487, 0, 0, 0, 0])\n",
        "ops.load(14, *[0, 0.2248230818150756, 0, 0, 0, 0])\n",
        "ops.load(19, *[0, 1.032752748769425, 0, 0, 0, 0])\n",
        "ops.load(20, *[0, 1.0327527487694255, 0, 0, 0, 0])\n",
        "ops.load(15, *[0, 0.22482308181507565, 0, 0, 0, 0])\n",
    ]

    def _ref_vals(s):
        start = s.find("[")
        return eval(s[start : s.find("]") + 1])

    for i, load_command in enumerate(example_bridge.load_case_list[0]["load_command"]):
        _name, args, _kw = load_command
        pos = list(args[1:])  # drop node_tag; rest are the 6 load components
        pos_ref = _ref_vals(ref_answer[i])
        assert pos == pytest.approx(pos_ref)


def test_local_vs_global_coord_settings():
    location = og.create_load_vertex(x=5, y=0, z=-2, p=20)  # create load point
    local_location = og.create_load_vertex(x=0, y=0, z=0, p=20)
    local_point_load = og.create_load(
        loadtype="point",
        name="single point",
        point1=local_location,
        shape_function="hermite",
    )  # defined for local coordinate
    global_point_load = og.create_load(
        loadtype="point", name="single point", point1=location, shape_function="hermite"
    )  # defined for local coordinate

    M1600_local = og.CompoundLoad("Truck model")
    M1600_local.add_load(
        load=local_point_load
    )  # if local_coord is set, append the local coordinate of the point load
    M1600_local.set_global_coord(og.Point(5, 0, -2))

    M1600_global = og.CompoundLoad("Truck model global")
    M1600_global.add_load(global_point_load)
    assert (
        M1600_local.compound_load_obj_list[0].load_point_1
        == M1600_global.compound_load_obj_list[0].load_point_1
    )


# test analysis of moving load case, test pass if no errors are returned
def test_moving_load_case(bridge_model_42_negative):
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    front_wheel = og.create_load(
        loadtype="point",
        name="front wheel",
        point1=og.LoadPoint(2, 0, 2, 50),
        shape_function="hermite",
    )  # Single point load 50 N

    single_path = og.create_moving_path(
        start_point=og.Point(2, 0, 2), end_point=og.Point(4, 0, 3)
    )  # create path object
    move_point = og.create_moving_load(name="single_moving_point")
    move_point.set_path(single_path)
    move_point.add_load(front_wheel)
    example_bridge.add_load_case(move_point)

    example_bridge.analyze()
    results = example_bridge.get_results()
    print(results)


# test moving load + basic load case
# test analysis of moving load case, test pass if no errors are returned
def test_moving_load_and_basic_load_together(bridge_model_42_negative):
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    # create reference line load
    barrierpoint_1 = og.create_load_vertex(x=5, y=0, z=1, p=2)
    barrierpoint_2 = og.create_load_vertex(x=10, y=0, z=1, p=2)
    Barrier = og.create_load(
        loadtype="line",
        name="Barrier curb load",
        point1=barrierpoint_1,
        point2=barrierpoint_2,
        shape_function="hermite",
    )
    barrier_load_case = og.create_load_case(name="Barrier")
    barrier_load_case.add_load(Barrier)  # ch
    example_bridge.add_load_case(barrier_load_case)
    # add moving load case
    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(2, 0, 2, 50), shape_function="hermite"
    )  # Single point load 50 N

    single_path = og.create_moving_path(
        start_point=og.Point(2, 0, 2), end_point=og.Point(4, 0, 3)
    )  # create path object
    move_point = og.create_moving_load(name="single_moving_point")
    move_point.set_path(single_path)
    move_point.add_load(front_wheel)
    example_bridge.add_load_case(move_point)

    # example_bridge.analyze(all=True)
    example_bridge.analyze()
    # results = example_bridge.get_results(load_case="single_moving_point")
    results = example_bridge.get_results(
        combinations={"Barrier": 1, "single_moving_point": 2}
    )
    print(results)


# test moving compound load, test pass if no errors are returned
def test_moving_compound_load(bridge_model_42_negative):
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    M1600 = og.CompoundLoad("M1600 LM")
    back_wheel = og.create_load(
        loadtype="point",
        name="single point",
        point1=og.LoadPoint(5, 0, 2, 20),
        shape_function="hermite",
    )  # Single point load 20 N
    front_wheel = og.create_load(
        loadtype="point",
        name="front wheel",
        point1=og.LoadPoint(2, 0, 2, 50),
        shape_function="hermite",
    )  # Single point load 50 N
    # compound the point loads
    M1600.add_load(back_wheel)
    M1600.add_load(front_wheel)
    M1600.set_global_coord(og.Point(0, 0, 0))

    truck = og.create_moving_load(name="Truck 1")
    single_path = og.create_moving_path(
        start_point=og.Point(2, 0, 2), end_point=og.Point(4, 0, 2)
    )  # Path object
    truck.set_path(path_obj=single_path)
    truck.add_load(M1600)

    example_bridge.add_load_case(truck)
    example_bridge.analyze()
    results = example_bridge.get_results()
    print(results)
    print("finish test compound moving load")


# checks if patch load is correctly-distributed when patch exceeds the bounds of the grillage
def test_patch_partially_outside_mesh(bridge_model_42_negative):
    example_bridge = bridge_model_42_negative
    lane_point_1 = og.create_load_vertex(x=-5, z=3, p=5)
    lane_point_2 = og.create_load_vertex(x=8, z=3, p=5)
    lane_point_3 = og.create_load_vertex(x=8, z=5, p=5)
    lane_point_4 = og.create_load_vertex(x=-5, z=5, p=5)
    Lane = og.create_load(
        loadtype="patch",
        name="Lane 1",
        point1=lane_point_1,
        point2=lane_point_2,
        point3=lane_point_3,
        point4=lane_point_4,
        shape_function="hermite",
    )
    ULS_DL = og.create_load_case(name="Lane")
    ULS_DL.add_load(Lane)  # ch
    example_bridge.add_load_case(ULS_DL)
    example_bridge.analyze()
    results = example_bridge.get_results()

    ref_answer = [
        "ops.load(10, *[0, 1.1724010993461413, 0, 0.0, 0, 0.0])\n",
        "ops.load(14, *[0, 1.1724010993461444, 0, 0.0, 0, 0.0])\n",
        "ops.load(15, *[0, 1.172401099346146, 0, 0.0, 0, 0.0])\n",
        "ops.load(14, *[0, 1.7586016490192202, 0, 0.8793008245096101, 0, 0.8793008245096101])\n",
        "ops.load(19, *[0, 1.7586016490192202, 0, 0.8793008245096101, 0, -0.8793008245096101])\n",
        "ops.load(20, *[0, 1.7586016490192202, 0, -0.8793008245096101, 0, -0.8793008245096101])\n",
        "ops.load(15, *[0, 1.7586016490192202, 0, -0.8793008245096101, 0, 0.8793008245096101])\n",
        "ops.load(19, *[0, 1.4068813192153762, 0, 0.7034406596076881, 0, 0.7034406596076881])\n",
        "ops.load(25, *[0, 1.4068813192153762, 0, 0.7034406596076881, 0, -0.7034406596076881])\n",
        "ops.load(26, *[0, 1.4068813192153762, 0, -0.7034406596076881, 0, -0.7034406596076881])\n",
        "ops.load(20, *[0, 1.4068813192153762, 0, -0.7034406596076881, 0, 0.7034406596076881])\n",
        "ops.load(25, *[0, 1.4442076913730961, 0, 0.7221038456865481, 0, 0.7221038456865481])\n",
        "ops.load(60, *[0, 1.4442076913730961, 0, 0.7221038456865481, 0, -0.7221038456865481])\n",
        "ops.load(61, *[0, 1.4442076913730961, 0, -0.7221038456865481, 0, -0.7221038456865481])\n",
        "ops.load(26, *[0, 1.4442076913730961, 0, -0.7221038456865481, 0, 0.7221038456865481])\n",
        "ops.load(6, *[0, 0.07503367035815302, 0, 0.0, 0, 0.0])\n",
        "ops.load(9, *[0, 0.07503367035815302, 0, 0.0, 0, 0.0])\n",
        "ops.load(10, *[0, 0.41268518696984174, 0, 0.0, 0, 0.0])\n",
        "ops.load(9, *[0, 0.14631565719839912, 0, 0.09004040442978405, 0, 0.07315782859919957])\n",
        "ops.load(13, *[0, 0.1463156571983994, 0, 0.09004040442978421, 0, -0.07315782859919967])\n",
        "ops.load(14, *[0, 1.2605656620169758, 0, -0.3601616177191363, 0, -0.6302828310084878])\n",
        "ops.load(10, *[0, 1.2605656620169736, 0, -0.3601616177191357, 0, 0.630282831008487])\n",
        "ops.load(13, *[0, 0.14631565719839917, 0, 0.09004040442978412, 0, 0.0731578285991996])\n",
        "ops.load(18, *[0, 0.1463156571983993, 0, 0.09004040442978421, 0, -0.07315782859919966])\n",
        "ops.load(19, *[0, 1.2605656620169776, 0, -0.3601616177191365, 0, -0.6302828310084888])\n",
        "ops.load(14, *[0, 1.2605656620169763, 0, -0.36016161771913613, 0, 0.6302828310084883])\n",
        "ops.load(18, *[0, 0.1170525257587192, 0, 0.07203232354382717, 0, 0.05852626287935966])\n",
        "ops.load(24, *[0, 0.11705252575871981, 0, 0.07203232354382755, 0, -0.05852626287935985])\n",
        "ops.load(25, *[0, 1.0084525296135833, 0, -0.28812929417530975, 0, -0.5042262648067912])\n",
        "ops.load(19, *[0, 1.008452529613578, 0, -0.28812929417530825, 0, 0.5042262648067894])\n",
        "ops.load(24, *[0, 0.12015807992224022, 0, 0.07394343379830169, 0, 0.06007903996112021])\n",
        "ops.load(59, *[0, 0.1201580799222415, 0, 0.07394343379830248, 0, -0.06007903996112065])\n",
        "ops.load(60, *[0, 1.035208073176236, 0, -0.2957737351932102, 0, -0.5176040365881172])\n",
        "ops.load(25, *[0, 1.0352080731762252, 0, -0.295773735193207, 0, 0.5176040365881135])\n",
        "ops.load(59, *[0, 0.12494235320970869, 0, 0.07688760197520536, 0, 0.056897774410875845])\n",
        "ops.load(66, *[0, 0.07600612632374003, 0, 0.04677300081460926, 0, -0.040880164010709555])\n",
        "ops.load(67, *[0, 0.6548220114045306, 0, -0.1870920032584372, 0, -0.35219833609226747])\n",
        "ops.load(60, *[0, 1.0764264276528765, 0, -0.3075504079008217, 0, 0.490196210309085])\n",
        "ops.load(60, *[0, 1.501710976078218, 0, 0.7508554880391092, 0, 0.6838674808999459])\n",
        "ops.load(67, *[0, 0.9135351721603383, 0, 0.4567675860801694, 0, -0.4913481251287209])\n",
        "ops.load(68, *[0, 0.9135351721603405, 0, -0.45676758608017004, 0, -0.49134812512872195])\n",
        "ops.load(61, *[0, 1.501710976078221, 0, -0.7508554880391103, 0, 0.6838674808999475])\n",
        "ops.load(61, *[0, 0.5838652274992154, 0, 0.09731087124986965, 0, 0.2658876765739])\n",
        "ops.load(68, *[0, 0.35518247493593585, 0, 0.059197079155989554, 0, -0.19103615105004512])\n",
        "ops.load(69, *[0, 0.01023159392819577, 0, -0.006577453239554427, 0, -0.005503099001441674])\n",
        "ops.load(62, *[0, 0.016819162932076304, 0, -0.010812319027763346, 0, 0.007659315786079486])\n",
        "ops.load(15, *[0, 0.5697869342822274, 0, 0.0, 0, 0.0])\n",
        "ops.load(20, *[0, 0.5697869342822279, 0, 0.0, 0, 0.0])\n",
        "ops.load(21, *[0, 0.12661931872938303, 0, 0.0, 0, 0.0])\n",
        "ops.load(20, *[0, 0.5469954569109349, 0, 0.09116590948515615, 0, 0.2734977284554669])\n",
        "ops.load(26, *[0, 0.5469954569109275, 0, 0.09116590948515492, 0, -0.27349772845546444])\n",
        "ops.load(27, *[0, 0.015757070775212, 0, -0.010129545498350583, 0, -0.007878535387606017])\n",
        "ops.load(21, *[0, 0.015757070775212212, 0, -0.01012954549835072, 0, 0.007878535387606089])\n",
        "ops.load(26, *[0, 0.5615079504058634, 0, 0.09358465840097817, 0, 0.2807539752029313])\n",
        "ops.load(61, *[0, 0.5615079504058589, 0, 0.09358465840097742, 0, -0.28075397520292983])\n",
        "ops.load(62, *[0, 0.01617512614337903, 0, -0.010398295377886516, 0, -0.008087563071689525])\n",
        "ops.load(27, *[0, 0.016175126143379157, 0, -0.0103982953778866, 0, 0.008087563071689568])\n",
    ]

    def _ref_vals(s):
        start = s.find("[")
        return eval(s[start : s.find("]") + 1])

    for i, load_command in enumerate(example_bridge.load_case_list[0]["load_command"]):
        _name, args, _kw = load_command
        pos = list(args[1:])  # drop node_tag; rest are the 6 load components
        pos_ref = _ref_vals(ref_answer[i])
        assert pos == pytest.approx(pos_ref)


def test_clearing_results(bridge_model_42_negative):
    # test functionality of clearing load case after analysis
    example_bridge = bridge_model_42_negative
    lane_point_1 = og.create_load_vertex(x=-5, z=3, p=5)
    lane_point_2 = og.create_load_vertex(x=8, z=3, p=5)
    lane_point_3 = og.create_load_vertex(x=8, z=5, p=5)
    lane_point_4 = og.create_load_vertex(x=-5, z=5, p=5)
    Lane = og.create_load(
        loadtype="patch",
        name="Lane 1",
        point1=lane_point_1,
        point2=lane_point_2,
        point3=lane_point_3,
        point4=lane_point_4,
        shape_function="hermite",
    )
    ULS_DL = og.create_load_case(name="Lane")
    ULS_DL.add_load(Lane)  # ch
    example_bridge.add_load_case(ULS_DL)
    example_bridge.analyze()
    results = example_bridge.get_results()
    assert example_bridge.results.basic_load_case_record != {}
    assert example_bridge.load_case_list != []
    example_bridge.clear_load_cases(load_case="Lane")
    assert example_bridge.results.basic_load_case_record == {}
    assert example_bridge.load_case_list == []


def test_load_analysis_shell_model(shell_link_bridge):
    # checks a benchmark analysis on a shell_beam model
    shell_link_model = shell_link_bridge
    # og.opsplt.plot_model("nodes")

    lane_point_1 = og.create_load_vertex(x=5, y=0, z=3, p=5)
    lane_point_2 = og.create_load_vertex(x=8, y=0, z=3, p=5)
    lane_point_3 = og.create_load_vertex(x=8, y=0, z=5, p=5)
    lane_point_4 = og.create_load_vertex(x=5, y=0, z=5, p=5)
    Lane = og.PatchLoading(
        point1=lane_point_1,
        point2=lane_point_2,
        point3=lane_point_3,
        point4=lane_point_4,
    )
    ULS_DL = og.LoadCase(name="Lane")
    ULS_DL.add_load(Lane)  # ch
    shell_link_model.add_load_case(ULS_DL)
    shell_link_model.analyze()

    results = shell_link_model.get_results()
    print(results)


def test_load_analysis_shell_multi_span(ref_bridge_properties):
    # checks integration of load analysis with multi - span feature + shell beam model
    def create_multispan_shell_model(ref_bridge_properties):
        # creates a reference shell beam model
        I_beam, slab, exterior_I_beam, concrete = ref_bridge_properties

        # Adopted units: N and m
        kilo = 1e3
        milli = 1e-3
        m = 1
        m2 = m**2
        m3 = m**3
        m4 = m**4

        # parameters of bridge grillage
        L = 30 * m  # span
        w = 10 * m  # width
        n_l = 7  # number of longitudinal members
        n_t = 11  # number of transverse members
        edge_dist = 1.05 * m  # distance between edge beam and first exterior beam
        bridge_name = "multi span showcase"
        angle = 20  # degree
        mesh_type = "Oblique"
        model_type = "shell_beam"
        # multispan specific vars
        spans = [9 * m, 12 * m, 9 * m]
        nl_multi = [20, 10, 20]
        stich_slab_x_spacing = 0.5 * m

        variant_one_model = og.create_grillage(
            bridge_name=bridge_name,
            long_dim=L,
            width=w,
            skew=angle,
            num_long_grid=n_l,
            num_trans_grid=n_t,
            edge_beam_dist=edge_dist,
            mesh_type=mesh_type,
            model_type=model_type,
            multi_span_dist_list=spans,
            multi_span_num_points=nl_multi,
            continuous=False,
            non_cont_spacing_x=stich_slab_x_spacing,
            max_mesh_size_z=1,
            max_mesh_size_x=1,
            offset_beam_y_dist=0.499,
            beam_width=0.89,
        )

        # create material of slab shell
        slab_shell_mat = og.create_material(
            material="concrete", code="AS5100-2017", grade="50MPa", rho=2400
        )

        # create section of slab shell
        slab_shell_section = og.create_section(h=0.2)
        slab_shell = og.create_member(
            section=slab_shell_section, material=slab_shell_mat
        )

        # create stitch slab conencting elements
        stitch_slab_section = og.create_section(
            A=0.504 * m2,
            J=5.22303e-3 * m3,
            Iy=0.32928 * m4,
            Iz=1.3608e-3 * m4,
            Ay=0.42 * m2,
            Az=0.42 * m2,
        )
        stitch_slab = og.create_member(section=stitch_slab_section, material=concrete)

        # set shell
        variant_one_model.set_shell_members(slab_shell)

        # assign grillage member to element groups of grillage model
        variant_one_model.set_member(I_beam, member="interior_main_beam")
        variant_one_model.set_member(I_beam, member="exterior_main_beam_1")
        variant_one_model.set_member(I_beam, member="exterior_main_beam_2")
        variant_one_model.set_member(exterior_I_beam, member="edge_beam")
        # variant_one_model.set_member(stitch_slab, member="stitch_elements")

        variant_one_model.create_osp_model(pyfile=False)

        return variant_one_model

    shell_bridge = create_multispan_shell_model(ref_bridge_properties)

    # create and add load case comprise of single point load
    P = 20e3
    point_load_location = og.create_load_vertex(
        x=4.5, y=0, z=6.5, p=P
    )  # about midspan of span 1
    point_load = og.create_load(
        loadtype="point", name="single point", point1=point_load_location
    )
    point_lc = og.create_load_case(name="pointload")
    point_lc.add_load(point_load)
    shell_bridge.add_load_case(point_lc)
    shell_bridge.analyze()
    # extract results
    result = shell_bridge.get_results()
    print(result)


def test_load_analysis_on_spring_support_single_span(ref_bridge_properties):
    # test multispan feature
    I_beam, slab, exterior_I_beam, concrete = ref_bridge_properties

    # parameters of bridge grillage
    L = 33.5 * m  # span
    w = 11.565 * m  # width
    n_l = 7  # number of longitudinal members
    n_t = 11  # number of transverse members
    edge_dist = 1.05 * m  # distance between edge beam and first exterior beam
    bridge_name = "multi span showcase"
    angle = 10  # degree
    mesh_type = "Oblique"

    # multispan specific vars
    spans = [9 * m, 12 * m, 9 * m]
    nl_multi = [3, 5, 10]
    stich_slab_x_spacing = 1 * m
    stitch_slab_section = og.create_section(
        A=0.504 * m2,
        J=5.22303e-3 * m3,
        Iy=0.32928 * m4,
        Iz=1.3608e-3 * m4,
        Ay=0.42 * m2,
        Az=0.42 * m2,
    )
    stich_slab = og.create_member(section=stitch_slab_section, material=concrete)

    variant_one_model = og.create_grillage(
        bridge_name=bridge_name,
        long_dim=L,
        width=w,
        skew=angle,
        num_long_grid=n_l,
        num_trans_grid=n_t,
        edge_beam_dist=edge_dist,
        mesh_type=mesh_type,
        multi_span_dist_list=spans,
        multi_span_num_points=nl_multi,
        continuous=True,
        # non_cont_spacing_x=stich_slab_x_spacing,
    )

    # assign grillage member to element groups of grillage model
    variant_one_model.set_member(I_beam, member="interior_main_beam")
    variant_one_model.set_member(I_beam, member="exterior_main_beam_1")
    variant_one_model.set_member(I_beam, member="exterior_main_beam_2")
    variant_one_model.set_member(exterior_I_beam, member="edge_beam")
    variant_one_model.set_member(slab, member="transverse_slab")
    variant_one_model.set_member(slab, member="start_edge")
    variant_one_model.set_member(slab, member="end_edge")
    # variant_one_model.set_member(stich_slab, member="stitch_elements")

    # spring support
    e_spring = 11e13
    variant_one_model.set_spring_support(
        rotational_spring_stiffness=e_spring, edge_num=1
    )

    variant_one_model.create_osp_model(pyfile=False)
    # og.opsplt.plot_model()

    # create and add load case comprise of single point load
    P = 20e3
    point_load_location = og.create_load_vertex(
        x=7.5, y=0, z=3, p=P
    )  # about midspan of span 1
    point_load = og.create_load(name="single point", point1=point_load_location)
    point_lc = og.create_load_case(name="pointload")
    point_lc.add_load(point_load)
    variant_one_model.add_load_case(point_lc)
    variant_one_model.analyze()

    result = variant_one_model.get_results()
    print(result)

    # Reference value updated: oblique mesh edge transverse elements are now
    # correctly classified as start_edge/end_edge instead of transverse_slab,
    # giving them proper edge section properties.
    ref_value = result.forces.sel(
        Loadcase="pointload", Component="Mz_i", Element=20
    ).to_numpy()
    assert og.np.isclose(
        ref_value.tolist(),
        -0.03730929768810762,
    )


def test_compare_shell_beam_analysis(run_beam_model_point_load):
    # run fixture to get bridges
    beam_bridge, result_beam, shell_bridge, result_shell = run_beam_model_point_load
    beam_bridge.get_element(member="exterior_main_beam_1")
    len(
        result_beam.forces.sel(
            Loadcase="line",
            Element=beam_bridge.get_element(member="exterior_main_beam_1")[0],
        )
    )
    # get middle interior beam results
    og.plot_force(
        beam_bridge, result_beam, member="exterior_main_beam_2", component="Mz"
    )
    og.plot_def(
        beam_bridge,
        result_beam,
        members="exterior_main_beam_1",
    )
    og.plot_def(
        shell_bridge,
        result_shell,
        members="interior_main_beam",
    )
    # og.opsv.plot_defo()  # removed: opsv is not part of the public API


def test_transient(beam_element_bridge):
    """A test for the uncoupled transient analysis portion of VBI basic framework"""
    positions = [5, 5, 10, 15]
    P = 100 * kN
    first_step = True
    previous_state = None

    # start moving vehicle through all positions
    for x in positions:
        # create load case
        mid_point_line_loadcase = og.create_load_case(name="VBI")
        lp1 = og.create_load_vertex(x=x, y=0, z=3.5, p=P)
        mid_point_line_load = og.create_load(
            name="unit load",
            point1=lp1,
        )
        mid_point_line_loadcase.add_load(mid_point_line_load)

        # create bridge instance
        beam_bridge = beam_element_bridge
        beam_bridge.create_osp_model()

        print(og.ops.eigen(2))
        og.ops.rayleigh(0.0, 0.0, 0.0, 2 * 0.02 / 4)

        # set previous state
        if not first_step:
            beam_bridge.set_previous_state(previous_state)
        first_step = False

        # add load case, run and store
        beam_bridge.add_load_case(mid_point_line_loadcase)
        beam_bridge.analyze(analysis_type="Transient", step=100, time_increment=0.1)
        previous_state = beam_bridge.store_state()
        result = beam_bridge.get_results()
        print(result)
        postprocessor = og.PostProcessor(beam_bridge, result)

        contactpoints = postprocessor.get_arbitrary_displacements(point=[5, 0, 3.5])
        # do something with results
        print(result)
        print("Next step")


# ─────────────────────────────────────────────────────────────────────────────
# create_load_vertex – error paths
# ─────────────────────────────────────────────────────────────────────────────


def test_create_load_vertex_missing_x_raises():
    """create_load_vertex must raise ValueError when x is absent."""
    with pytest.raises(ValueError):
        og.create_load_vertex(z=1.0, p=10)


def test_create_load_vertex_missing_z_raises():
    """create_load_vertex must raise ValueError when z is absent."""
    with pytest.raises(ValueError):
        og.create_load_vertex(x=1.0, p=10)


def test_create_load_vertex_missing_both_raises():
    """create_load_vertex must raise ValueError when both x and z are absent."""
    with pytest.raises(ValueError):
        og.create_load_vertex(y=0, p=10)


# ─────────────────────────────────────────────────────────────────────────────
# create_load – nodal branch and error branch
# ─────────────────────────────────────────────────────────────────────────────


def test_create_load_nodal_returns_nodal_load():
    """create_load with loadtype='nodal' must return a NodalLoad."""
    load = og.create_load(
        loadtype="nodal",
        node_tag=42,
        Fx=0,
        Fy=-10e3,
        Fz=0,
        Mx=0,
        My=0,
        Mz=0,
        name="test_nodal",
    )
    assert isinstance(load, og.NodalLoad)


def test_create_load_nodal_stores_forces():
    """NodalLoad forces must match the values passed to create_load."""
    load = og.create_load(
        loadtype="nodal",
        node_tag=1,
        Fx=1.0,
        Fy=2.0,
        Fz=3.0,
        Mx=4.0,
        My=5.0,
        Mz=6.0,
    )
    assert load.Fx == 1.0
    assert load.Fy == 2.0
    assert load.Fz == 3.0
    assert load.Mx == 4.0
    assert load.My == 5.0
    assert load.Mz == 6.0


def test_create_load_invalid_type_raises():
    """create_load with no points and unknown loadtype must raise TypeError."""
    with pytest.raises(TypeError):
        og.create_load(loadtype="invalid_type_xyz")


# ─────────────────────────────────────────────────────────────────────────────
# NodalLoad – methods
# ─────────────────────────────────────────────────────────────────────────────


def test_nodal_load_get_nodal_load_call():
    """get_nodal_load_call must return the correct ('load', args, {}) tuple."""
    forces = og.NodeForces(Fx=0, Fy=-5e3, Fz=0, Mx=0, My=0, Mz=0)
    nl = og.NodalLoad(node_tag=7, node_force=forces, name="nl_test")
    func_name, args, kwargs = nl.get_nodal_load_call()
    assert func_name == "load"
    assert args[0] == 7  # node tag
    assert args[2] == -5e3  # Fy
    assert kwargs == {}


def test_nodal_load_get_nodal_load_str():
    """get_nodal_load_str must contain the node tag and force value."""
    forces = og.NodeForces(Fx=0, Fy=-1e3, Fz=0, Mx=0, My=0, Mz=0)
    nl = og.NodalLoad(node_tag=3, node_force=forces)
    s = nl.get_nodal_load_str()
    assert "ops.load(" in s
    assert "3" in s


def test_nodal_load_iterable_node_tags():
    """NodalLoad must accept a list of node tags and store one spec entry each."""
    forces = og.NodeForces(Fx=0, Fy=-1.0, Fz=0, Mx=0, My=0, Mz=0)
    nl = og.NodalLoad(node_tag=[10, 11, 12], node_force=forces)
    assert 10 in nl.spec
    assert 11 in nl.spec
    assert 12 in nl.spec


# ─────────────────────────────────────────────────────────────────────────────
# Loads.apply_load_factor() and get_magnitude()
# ─────────────────────────────────────────────────────────────────────────────


def test_apply_load_factor_to_point_load():
    """apply_load_factor() must multiply load_point_1.p by the given factor."""
    v = og.create_load_vertex(x=1, z=2, p=10.0)
    pt = og.create_load(loadtype="point", name="pt", point1=v)
    pt.apply_load_factor(2.0)
    assert pt.load_point_1.p == pytest.approx(20.0)


def test_apply_load_factor_to_all_line_points():
    """apply_load_factor() must scale both endpoints of a LineLoading."""
    v1 = og.create_load_vertex(x=0, z=0, p=5.0)
    v2 = og.create_load_vertex(x=0, z=4, p=5.0)
    line = og.create_load(loadtype="line", name="barrier", point1=v1, point2=v2)
    line.apply_load_factor(3.0)
    assert line.load_point_1.p == pytest.approx(15.0)
    assert line.load_point_2.p == pytest.approx(15.0)


def test_get_magnitude_returns_list_of_p_values():
    """get_magnitude() must return p values for all defined (non-None) load points."""
    v1 = og.create_load_vertex(x=0, z=0, p=7.0)
    v2 = og.create_load_vertex(x=0, z=4, p=9.0)
    line = og.create_load(loadtype="line", name="test_line", point1=v1, point2=v2)
    mags = line.get_magnitude()
    assert 7.0 in mags
    assert 9.0 in mags
    assert len(mags) == 2


# ─────────────────────────────────────────────────────────────────────────────
# LineLoading – curved (3-point) constructor
# ─────────────────────────────────────────────────────────────────────────────


def test_line_loading_curved_three_points():
    """LineLoading with a third midpoint must initialise without error."""
    p1 = og.create_load_vertex(x=0, z=0, p=1.0)
    p2 = og.create_load_vertex(x=5, z=2, p=1.0)
    p3 = og.create_load_vertex(x=10, z=0, p=1.0)
    curved = og.LineLoading(point1=p1, point2=p3, point3=p2, name="curved")
    assert hasattr(curved, "d")


# ─────────────────────────────────────────────────────────────────────────────
# PatchLoading – 4-point quadrilateral
# ─────────────────────────────────────────────────────────────────────────────


def test_patch_load_four_points():
    """PatchLoading with four CCW-ordered vertices must initialise without error."""
    p1 = og.create_load_vertex(x=0, z=0, p=10.0)
    p2 = og.create_load_vertex(x=5, z=0, p=10.0)
    p3 = og.create_load_vertex(x=5, z=4, p=10.0)
    p4 = og.create_load_vertex(x=0, z=4, p=10.0)
    patch = og.PatchLoading(
        point1=p1, point2=p2, point3=p3, point4=p4, name="quad_patch"
    )
    assert patch.load_point_1 is not None
    assert patch.load_point_4 is not None


# ─────────────────────────────────────────────────────────────────────────────
# Path – get_path_points and get_custom_path_points
# ─────────────────────────────────────────────────────────────────────────────


def test_path_get_path_points_returns_correct_length():
    """get_path_points must return a list of length equal to increments."""
    start = og.Point(0, 0, 0)
    end = og.Point(10, 0, 0)
    p = og.Path(start_point=start, end_point=end, increments=10)
    pts = p.get_path_points()
    assert len(pts) == 10


def test_path_get_custom_path_points_returns_correct_length():
    """get_custom_path_points must return a list of length new_increment."""
    start = og.Point(0, 0, 0)
    end = og.Point(10, 0, 0)
    p = og.Path(start_point=start, end_point=end, increments=20)
    pts = p.get_custom_path_points(new_increment=5)
    assert len(pts) == 5


# ─────────────────────────────────────────────────────────────────────────────
# LoadModel – M1600 vehicle
# ─────────────────────────────────────────────────────────────────────────────


def test_load_model_m1600_creates_compound_load():
    """LoadModel.create() for M1600 must return a CompoundLoad with 24 point loads."""
    lm = og.LoadModel(model_type="M1600", gap=5.0)
    vehicle = lm.create()
    assert vehicle is not None
    assert len(vehicle.compound_load_obj_list) == 24


def test_load_model_m1600_non_zero_loads():
    """All M1600 point loads must have a non-zero p value."""
    lm = og.LoadModel(model_type="M1600", gap=5.0)
    vehicle = lm.create()
    for point_load in vehicle.compound_load_obj_list:
        assert point_load.load_point_1.p > 0


# ─────────────────────────────────────────────────────────────────────────────
# MovingLoad – error path when no path is set
# ─────────────────────────────────────────────────────────────────────────────


def test_moving_load_add_load_no_path_raises():
    """MovingLoad.add_load() without a path and no set_path must raise ValueError."""
    ml = og.create_moving_load(name="test_ml")
    v = og.create_load_vertex(x=0, z=0, p=10.0)
    pt = og.create_load(loadtype="point", name="pt", point1=v)
    lc = og.create_load_case(name="lc_for_ml")
    lc.add_load(pt)
    with pytest.raises(ValueError):
        ml.add_load(lc)
        ml.parse_moving_load_cases()


# ─────────────────────────────────────────────────────────────────────────────
# ShapeFunction
# ─────────────────────────────────────────────────────────────────────────────


def test_shape_function_get_shape_function_hermite():
    """ShapeFunction.get_shape_function('hermite') must return a callable."""
    sf = og.ShapeFunction()
    fn = sf.get_shape_function(eta=0.0, zeta=0.0, option="hermite")
    assert callable(fn)


def test_shape_function_get_shape_function_linear():
    """ShapeFunction.get_shape_function('linear') must return a callable."""
    sf = og.ShapeFunction()
    fn = sf.get_shape_function(eta=0.5, zeta=0.5, option="linear")
    assert callable(fn)


def test_shape_function_hermite_1d_length():
    """hermite_shape_function_1d must return four terms."""
    result = og.ShapeFunction.hermite_shape_function_1d(zeta=0.5, a=1.0)
    assert len(result) == 4



# -*- coding: utf-8 -*-
"""
This module manages user interface functions and class definitions for
grillage members. In our terminology, we define a section as the geometrical
properties of the structural elements, and the member as the combination of
the section and material properties.
"""
from typing import TYPE_CHECKING

# if TYPE_CHECKING:
#     from material import Material
g = 9.81
mm = 1e3

__all__ = ["Section", "GrillageMember", "create_section", "create_member"]


def create_section(**kwargs):
    """
    Create a section object defining cross-section geometric properties.

    The constructor :class:`Section` takes the following arguments.

    :param op_ele_type: OpenSees element type - default is "elasticBeamColumn"
    :type op_ele_type: str
    :param op_section_type: OpenSees section type - default is "Elastic"
    :type op_section_type: str
    :param unit_width: Flag for unit width properties
    :type unit_width: bool

    The constructor of :class:`Section` takes the following ``kwargs``:

    :param A: Cross-sectional area.
    :type A: float, optional
    :param Iz: Second moment of area about the local z axis.
    :type Iz: float, optional
    :param Iy: Second moment of area about the local y axis.
    :type Iy: float, optional
    :param J: Torsional inertia about the local x axis.
    :type J: float, optional
    :param Az: Shear area in the local z direction.
    :type Az: float, optional
    :param Ay: Shear area in the local y direction.
    :type Ay: float, optional
    :param offset_y: Vertical offset from section centroid to the grillage model plane.
        When provided, the parallel axis theorem is applied:
        ``I_offset = I_centroid + A * offset_y**2`` for both Iz and Iy.
        J (torsion) is not modified.
    :type offset_y: float, optional
    :returns: :class:`Section` object.
    """
    return Section(**kwargs)


def create_member(**kwargs):
    """
    Create a grillage member combining a section and a material.

    A grillage member requires a :class:`~ospgrillage.material.Material` and a
    :class:`~ospgrillage.members.Section`.

    :param section: Section class object
    :type section: :class:`Section`
    :param material: Material class object
    :type material: :class:`~ospgrillage.material.Material`
    :param member_name: Name of the grillage member (Optional)
    :type member_name: str

    :returns: :class:`~ospgrillage.members.GrillageMember` object
    """
    return GrillageMember(**kwargs)


class Section:
    """
    Class for structural sections of grillage model. Stores geometric properties of cross sections. This class also
    parses section inputs into relevant
    ``OpenSeesPy`` command for creating sections in OpenSees framework.

    This class wraps ```OpenSeesPy``` Section() commands - methods for generating commands are handled by the
    higher hierarchy:class:`~ospgrillage.member.GrillageMember` class

    """

    def __init__(
        self,
        op_ele_type: str = "elasticBeamColumn",
        mass: float = 0,
        c_mass_flag: bool = False,
        unit_width: bool = False,
        op_section_type: str = "Elastic",
        offset_y: float = None,
        **kwargs,
    ):
        """
        The constructor takes in two types of keyword arguments.

        #. General section properties - such as A, I, J for example. These properties are parses into the appropriate
           OpenSees section arguments.
        #. ``OpenSeespy`` section arguments - i.e. specific keyword for a specific ops.section() type.

        `Section <https://openseespydoc.readthedocs.io/en/latest/src/section.html>`_  information of OpenSeesPy

        Constructor takes following inputs to be parse into arguments for ```OpenSeesPy``` Sections.

        :param op_ele_type: OpenSees element type - default elasticBeamColumn
        :type op_ele_type: str
        :param op_section_type: OpenSees section type - default Elastic
        :type op_section_type: str
        :param mass: Mass of member.
        :type mass: float
        :param unit_width: Flag for if unit width properties are defined.
        :type unit_width: bool

        Constructor also takes the following keyword arguments for a section.

        :param A: Cross-sectional area.
        :type A: float, optional
        :param Iz: Second moment of area about the local z axis.
        :type Iz: float, optional
        :param Iy: Second moment of area about the local y axis.
        :type Iy: float, optional
        :param J: Torsional inertia about the local x axis.
        :type J: float, optional
        :param Az: Shear area in the local z direction.
        :type Az: float, optional
        :param Ay: Shear area in the local y direction.
        :type Ay: float, optional
        :param offset_y: Vertical offset from section centroid to the grillage model
            plane. When provided, the parallel axis theorem is applied:
            ``I_offset = I_centroid + A * offset_y**2`` for both Iz and Iy.
            J (torsion) is not modified.
        :type offset_y: float, optional

        .. note::

            Section properties are defined in local coordinates (x, y, z) and are later
            transformed according to the orientation of the :class:`GrillageMember` in the grillage model.
        """
        # OpenSees py section type
        self.op_section_type = (
            op_section_type  # section tag based on OpenSeesPy - default is elastic
        )
        self.section_command_flag = False  # flag for parsing , check if section() command is needed. default False
        # OpenSeesPy section arguments
        # standard geometric properties
        self.A = kwargs.get("A", None)
        self.Iz = kwargs.get("Iz", None)
        self.Iy = kwargs.get("Iy", None)
        # for 3D sections
        self.Ay = kwargs.get("Ay", None)
        self.Az = kwargs.get("Az", None)
        self.J = kwargs.get("J", None)
        self.offset_y = offset_y
        # for non-linear properties
        self.alpha_y = kwargs.get("alpha_y", None)
        self.alpha_z = kwargs.get("alpha_z", None)
        self.K11 = kwargs.get("K11", None)
        self.K33 = kwargs.get("K33", None)
        self.K44 = kwargs.get("K44", None)
        # for quad elements
        self.h_depth = kwargs.get("h", None)

        # types for element definition and unit width properties
        self.op_ele_type = op_ele_type
        self.unit_width = unit_width
        self.mass = mass
        self.c_mass_flag = c_mass_flag
        # keyword args
        self.num_int_pt = kwargs.get("num_int_pt", None)
        self.integration_type = kwargs.get("integration_type", None)
        # quad/tri element parameters
        self.thick = kwargs.get("thick", None)
        # for quad elements
        # here check and overwrite element if  h_depth is given, default to quad shell element
        if self.h_depth:
            self.op_section_type = "ElasticMembranePlateSection"
            self.op_ele_type = "ShellDKGQ"
        # warning checks
        E = kwargs.get("E", None)
        G = kwargs.get("G", None)
        if any([E, G]):
            raise ValueError(
                "E or G is provided to Section object. Hint: E and G are attributes of Material Object"
            )
        self.parse_section_properties()

    def parse_section_properties(self):
        """
        Parse and populate missing section properties with computed defaults.

        This method computes secondary properties (Ay, Az, Iy) based on primary
        section properties (A, Iz) if not explicitly provided. These computed
        values are required for OpenSees model commands.
        """
        # function to parse input properties
        # mainly, this is used to define less essential properties (e.g. Ay, Az) which user may not specify
        # but is required for the model commands
        # for developers, add here the section properties which needs to be parse for model command
        if self.Ay is None and self.A is not None:
            self.Ay = 0.5 * self.A
        if self.Az is None and self.A is not None:
            self.Az = 0.2 * self.A
        if self.Iy is None and self.Iz is not None:
            self.Iy = 0.2 * self.Iz

        # Apply parallel axis theorem when offset_y is provided
        if self.offset_y is not None and self.A is not None:
            d_sq = self.offset_y**2
            if self.Iz is not None:
                self.Iz += self.A * d_sq
            if self.Iy is not None:
                self.Iy += self.A * d_sq


# ----------------------------------------------------------------------------------------------------------------


class GrillageMember:
    """
    Class for grillage model members. E.g., longitudinal beam members.

    This class parses material and section properties of grillage members into corresponding ``OpenSeesPy`` Element()
    command.

    `Refer here <https://openseespydoc.readthedocs.io/en/latest/src/element.html>`_ for more information about
     ``OpenSeesPy`` Element types and definition workflow.

    For developers wishing to expand the library of elements wrapped by GrillageMember, introduce in this class:

    #. The argument input orders of the new element in
       :func:`~ospgrillage.member.GrillageMember.get_member_prop_arguments`.
    #. If the element requires an ``OpenSeesPy`` Section definition, add its input arguments to
       :func:`~ospgrillage.member.GrillageMember.get_section_arguments` and
       :func:`~ospgrillage.member.GrillageMember.get_ops_section_command`.

    """

    def __init__(
        self,
        section: Section,
        material,
        member_name: str = "Undefined",
        quad_ele_flag: bool = False,
        tri_ele_flag: bool = False,
    ):
        """
        Init the GrillageMember. Requires two input objects i.e. A :class:`~ospgrillage.material.Material`, and
        :class:`Section`.

        :param section: Section class object
        :type section: :class:`~ospgrillage.members.Section`
        :param material: Material class object
        :type material: :class:`~ospgrillage.material.Material`
        :param member_name: Name of the grillage member (Optional)
        :type member_name: str
        """
        self.member_name = member_name
        self.section = section
        self.material = material
        self.quad_flag = quad_ele_flag
        self.tri_ele_flag = tri_ele_flag
        self.section_command_flag = True
        self.material_command_flag = True
        if any(
            [
                self.section.op_ele_type == "ElasticTimoshenkoBeam",
                self.section.op_ele_type == "elasticBeamColumn",
            ]
        ):
            self.material_command_flag = False
            self.section_command_flag = False  #

            # determine mass from section area and material density
            self.mass = self.section.A * self.material.density * mm  # kg/mm
        elif any(
            [
                self.section.op_ele_type == "ShellMITC4",
                self.section.op_ele_type == "ShellDKGQ",
            ]
        ):
            self.material_command_flag = False

        self.variable_string_list = []

    def get_member_prop_arguments(self, width=1):
        """
        Returns the element arguments based on the op_element_type of the GrillageMember.

        :returns: str containing member properties in accordance with convention of OpenSees element type
        """

        asterisk_input = None

        # if elastic Beam column elements, return str of section input
        if self.section.op_ele_type == "ElasticTimoshenkoBeam":
            if None in [
                self.material.elastic_modulus,
                self.material.shear_modulus,
                self.section.A,
                self.section.J,
                self.section.Iy,
                self.section.Iz,
                self.section.Ay,
                self.section.Az,
            ]:
                raise ValueError(
                    "One or more missing arguments for Section: {}".format(
                        self.section.op_ele_type
                    )
                )
            asterisk_input = "[{:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}]".format(
                self.material.elastic_modulus,
                self.material.shear_modulus,
                self.section.A * width,
                self.section.J * width,
                self.section.Iy * width,
                self.section.Iz * width,
                self.section.Ay * width,
                self.section.Az * width,
            )
        elif self.section.op_ele_type == "elasticBeamColumn":  # eleColumn
            if None in [
                self.material.elastic_modulus,
                self.material.shear_modulus,
                self.section.A,
                self.section.J,
                self.section.Iy,
                self.section.Iz,
            ]:
                raise ValueError(
                    "One or more missing arguments for Section: {}".format(
                        self.section.op_ele_type
                    )
                )

            asterisk_input = "[{:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}]".format(
                self.section.A * width,
                self.material.elastic_modulus,
                self.material.shear_modulus,
                self.section.J * width,
                self.section.Iy * width,
                self.section.Iz * width,
            )

        elif self.section.op_ele_type == "ModElasticBeam2d":
            if None in [
                self.section.A,
                self.material.elastic_modulus,
                self.section.Iz * width,
                self.section.K11,
                self.section.K33,
                self.section.K44,
            ]:
                raise ValueError(
                    "One or more missing arguments for Section: {}".format(
                        self.section.op_section_type
                    )
                )
            asterisk_input = "[{:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}, {:.3e}]".format(
                self.section.A * width,
                self.material.elastic_modulus,
                self.section.Iz * width,
                self.section.K11,
                self.section.K33,
                self.section.K44,
            )

        # TO be populated with more inputs for various element types

        return asterisk_input

    # Function to return argument, handled by OspGrillage
    def get_section_arguments(self, ele_width: float = 1):
        """
        Returns the input arguments for ``OpenSeesPy`` Section command if the prescribe element type of the
        GrillageMember requires the definition of an ``OpenSeesPy`` Section object.

        :returns: str containing Section command properties in accordance with convention of OpenSees element type
        """
        section_args = None

        if self.section.op_section_type == "Elastic":
            section_args = [
                self.material.elastic_modulus,
                self.section.A,
                self.section.Iz,
                self.section.Iy,
                self.material.shear_modulus,
                self.section.J,
                self.section.alpha_y,
                self.section.alpha_z,
            ]

        return section_args

    def get_ops_section_command(
        self, section_tag: int = 1, material_tag: int = None, ele_width: float = 1
    ):
        """
        Returns the ``OpenSeesPy`` Section command for the Section type

        :returns: str containing the ``OpenSeesPy`` Section command
        """
        sec_str = None
        section_type = self.section.op_section_type
        if section_type == "Elastic":
            # section type, section tag, argument entries from self.get_asterisk_input()
            sec_arg = self.get_section_arguments(ele_width=ele_width)
            sec_str = 'ops.section("{type}", {tag}, *{arg})\n'.format(
                type=section_type, tag=section_tag, arg=repr(sec_arg)
            )
        elif section_type == "ElasticMembranePlateSection":
            # section type, section tag, E_mod, nu, h, rho
            sec_str = (
                'ops.section("{type}", {tag}, {E_mod}, {nu}, {h}, {rho})\n'.format(
                    type=section_type,
                    tag=section_tag,
                    E_mod=self.material.elastic_modulus,
                    nu=self.material.poisson_ratio,
                    h=self.section.h_depth,
                    rho=self.material.density,
                )
            )
        elif section_type == "PlateFiber":
            # section type, section tag, E_mod, nu, h, rho
            sec_str = 'ops.section("{type}", {tag}, {mat_tag}, {h})\n'.format(
                type=section_type,
                tag=section_tag,
                h=self.section.h_depth,
                mat_tag=material_tag,
            )

        return sec_str

    def get_element_command_str(
        self,
        ele_tag: int,
        node_tag_list: list,
        transf_tag: int = None,
        ele_width: float = 1,
        materialtag: int = None,
        sectiontag: int = None,
    ) -> str:
        """Return a list of OpenSeesPy element command for the member.
        This function is handled by OspGrillage class.
        """
        # ```
        # Function called within OpsGrillage class `set_member()` function.
        #
        # For shell elements, n1 n2 n3 n4 are counter clockwise node (n1 being node in quadrant -1 , -1 ).
        #
        # Procedure to be called
        # 1) OpsGrillage assigns the material and section first, then returns the tag of material and section
        # 2) OpsGrillage calls get_element_command_str of GrillageMember, then it takes in material section and returns
        #  the element command ops.element() for the respective grillage member

        # format of each ele sublist
        # [node i, node j, ele group, ele tag, transtag]
        # ```
        section_input = None
        ele_str = None
        if self.section.op_ele_type == "ElasticTimoshenkoBeam":
            section_input = self.get_member_prop_arguments(ele_width)
            ele_str = 'ops.element("{type}", {tag}, *{node_tag_list}, *{memberprop}, {transftag},"-mass", {mass})\n'.format(
                type=self.section.op_ele_type,
                tag=ele_tag,
                node_tag_list=node_tag_list,
                memberprop=section_input,
                transftag=transf_tag,
                mass=self.mass,
            )
        elif self.section.op_ele_type == "elasticBeamColumn":
            section_input = self.get_member_prop_arguments(ele_width)
            ele_str = 'ops.element("{type}", {tag}, *{node_tag_list}, *{memberprop}, {transftag}, "-mass", {mass})\n'.format(
                type=self.section.op_ele_type,
                tag=ele_tag,
                node_tag_list=node_tag_list,
                memberprop=section_input,
                transftag=transf_tag,
                mass=self.mass,
            )
        elif self.section.op_ele_type == "nonlinearBeamColumn":
            ele_str = 'ops.element("{type}",{tag},*{node_tag_list},{num_int_pt},{sectag},{transftag},"-mass", {mass})\n'.format(
                type=self.section.op_ele_type,
                tag=ele_tag,
                node_tag_list=node_tag_list,
                num_int_pt=self.section.num_int_pt,
                sectag=sectiontag,
                transftag=transf_tag,
                mass=self.mass,
            )
        elif self.section.op_ele_type == "zeroLength":
            ele_str = "ops.element(\"{linktype}\",{ele_tag},*[{rNodetag},{cNodetag}],'-mat',{mat_tag},'-dir',{dirs})\n".format(
                linktype=self.section.op_ele_type,
                rNodetag=node_tag_list[0],
                cNodetag=node_tag_list[1],
                dirs=6,
                mat_tag=materialtag,
                ele_tag=ele_tag,
            )

        # for shell element option
        elif self.section.op_ele_type == "ShellMITC4":
            ele_str = (
                'ops.element("{type}", {tag}, *{node_tag_list}, {sectag})\n'.format(
                    type=self.section.op_ele_type,
                    tag=ele_tag,
                    node_tag_list=node_tag_list,
                    sectag=sectiontag,
                )
            )

        elif self.section.op_ele_type == "ShellDKGQ":
            if len(node_tag_list) == 3:
                ele_type = "ShellDKGT"
            else:  # 4 node
                ele_type = "ShellDKGQ"
            ele_str = (
                'ops.element("{type}", {tag}, *{node_tag_list}, {sectag})\n'.format(
                    type=ele_type,
                    tag=ele_tag,
                    node_tag_list=node_tag_list,
                    sectag=sectiontag,
                )
            )

        elif self.section.op_ele_type == "ShellDKGT":
            ele_str = (
                'ops.element("{type}", {tag}, *{node_tag_list}, {sectag})\n'.format(
                    type=self.section.op_ele_type,
                    tag=ele_tag,
                    node_tag_list=node_tag_list,
                    sectag=sectiontag,
                )
            )

        return ele_str
